/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../common/square.h"
#include"cposition.h"
#include"../movegen/movegenerator.h"
#include"../common/prebb.h"
#include<algorithm>
#include<stdexcept>
#include<iostream>
#include<iomanip>
#include<sstream>
#include<chrono>
#include<random>
#include<cassert>
using std::istringstream;
using std::ostringstream;
using std::to_string;
using std::hex;
using std::dec;
using std::cout;
using std::cin;
using std::endl;
using std::logic_error;
using std::boolalpha;
using std::default_random_engine;
using std::uniform_int_distribution;
using namespace std::chrono;


int ChessPosition::cp_id=0;

ChessPosition::ChessPosition(const Tables& x) : m_tables(x), m_atta(x.attacks()), m_con(x.connections()), m_hh(Hasher<1>(x))
{
  m_cp_id=++cp_id;
  clear();
  //std::cerr << "ChessPosition()::ChessPosition() " << m_cp_id << endl;
};

ChessPosition::~ChessPosition()
{
  if(m_log){
    m_ofs.close();
  }
  //std::cerr << "ChessPosition::~ChessPosition() " << m_cp_id << endl;
};

ChessPosition::ChessPosition(const ChessPosition& x) : m_tables(x.m_tables), m_atta(x.m_atta), m_con(x.m_con), m_hh(Hasher<1>(m_tables))
{
  m_cp_id=++cp_id;
  m_make_move_counter=x.m_make_move_counter;
  fen(x.fen());
  m_hist=x.m_hist;
  m_initial_position=x.m_initial_position;
  m_rtf=x.m_rtf;
  m_log=x.m_log;
  //std::cerr << "ChessPosition::ChessPosition(const ChessPosition&) " << m_cp_id << endl;
}

ChessPosition& ChessPosition::operator=(const ChessPosition& x)
{
  if(&x!=this){
    m_make_move_counter=x.m_make_move_counter;
    fen(x.fen());
    m_hist=x.m_hist;
    m_initial_position=x.m_initial_position;
    m_rtf=x.m_rtf;
    m_log=x.m_log;
  }
  return *this;
}

void ChessPosition::clear()
{
  bp.clear();
  Piece p;
  m_initial_position="8/8/8/8/8/8/8/8 w - - 0 1";

  m_hist.clear();
  m_hist.push_back({BitBoard(),BitBoard(),BitBoard(),BitBoard(),0,0,0,NullMove(),no_square,no_square,no_square,false});
  
  m_total_moves=1;
  m_hh.init(*this);
  m_rtf.init(m_hh.hash());
  m_make_move_counter=0;
  compute_materialindex();
};

void ChessPosition::newgame()
{
  fen("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
  m_initial_position=fen();
};

void ChessPosition::newgame960(int opening)
{
  system_clock::time_point t1,t2{system_clock::now()};
  int seed=duration_cast<milliseconds>(t2-t1).count();
  default_random_engine dre{seed};
  uniform_int_distribution<int> uid(0,959);
  int nbr=uid(dre);
  if(opening>=0 && opening<960){nbr=opening;}
  
  Piece p;
  clear();

  int i=nbr/240;nbr%=240;
  p.set_white_bishop();
  switch(i){
  case 0:set_piece_on(a1,p);break;
  case 1:set_piece_on(c1,p);break;
  case 2:set_piece_on(e1,p);break;
  case 3:set_piece_on(g1,p);break;
  }
  i=nbr/60;nbr%=60;
  switch(i){
  case 0:set_piece_on(b1,p);break;
  case 1:set_piece_on(d1,p);break;
  case 2:set_piece_on(f1,p);break;
  case 3:set_piece_on(h1,p);break;
  }
  i=nbr/10;nbr%=10;
  p.set_white_queen();
  int j{0};
  for(Square s=a1;s<=h1;s++){
    if(get_piece_from(s).no_piece()){
      if(i==j){set_piece_on(s,p);break;}
      j++;
    }
  }

  int n1,n2;
  switch(nbr){
  case 0:n1=0,n2=1;break;
  case 1:n1=0,n2=2;break;
  case 2:n1=0,n2=3;break;
  case 3:n1=0,n2=4;break;
  case 4:n1=1,n2=2;break;
  case 5:n1=1,n2=3;break;
  case 6:n1=1,n2=4;break;
  case 7:n1=2,n2=3;break;
  case 8:n1=2,n2=4;break;
  case 9:n1=3,n2=4;break;
  default: throw logic_error("ChessPosition::newgame960");
  }

  p.set_white_knight();
  i=n1;
  j=0;
  for(Square s=a1;s<=h1;s++){
    if(get_piece_from(s).no_piece()){
      if(i==j){set_piece_on(s,p);break;}
      j++;
    }
  }
  i=n2;
  j=0;
  for(Square s=a1;s<=h1;s++){
    if(get_piece_from(s).no_piece()){
      if(i==j){set_piece_on(s,p);break;}
      j++;
    }
  }
  p.set_white_rook();
  for(Square s=a1;s<=h1;s++){if(get_piece_from(s).no_piece()){set_piece_on(s,p);break;}}
  p.set_white_king();
  for(Square s=a1;s<=h1;s++){if(get_piece_from(s).no_piece()){set_piece_on(s,p);break;}}
  p.set_white_rook();
  for(Square s=a1;s<=h1;s++){if(get_piece_from(s).no_piece()){set_piece_on(s,p);break;}}
  
  for(Square s=a1;s<=h1;s++){
    if(get_piece_from(s).white_knight()){p.set_black_knight();set_piece_on(s+56,p);}
    if(get_piece_from(s).white_bishop()){p.set_black_bishop();set_piece_on(s+56,p);}
    if(get_piece_from(s).white_rook()){p.set_black_rook();set_piece_on(s+56,p);}
    if(get_piece_from(s).white_queen()){p.set_black_queen();set_piece_on(s+56,p);}
    if(get_piece_from(s).white_king()){p.set_black_king();set_piece_on(s+56,p);}
  }
  
  p.set_white_pawn();for(Square s=a2;s<=h2;s++){set_piece_on(s,p);}
  p.set_black_pawn();for(Square s=a7;s<=h7;s++){set_piece_on(s,p);}

  uint16_t castlingrights=0;
  for(Square s=a1;s<=h1;s++){
    if(get_piece_from(s).white_rook()){
      castlingrights|=(1<<s.id());
    }
  }
  castlingrights|=castlingrights<<8;

  history_t h{BitBoard(),BitBoard(),BitBoard(),BitBoard(),m_make_move_counter,castlingrights,0,NullMove(),no_square,no_square,no_square,false};
  m_hist.back()=h;
  

  m_total_moves=1;

  m_hh.init(*this);
  m_initial_position=fen();
  fen(m_initial_position);
};

Move ChessPosition::convert_to_move(const string& x)
{
  string E="ChessPosition::convert_to_move(const string& " + x + ")";
  if(x.size()>5 || x.size()<4){throw logic_error(E);}
  if(x[0]<'a' || x[0]>'h'){throw logic_error(E);}
  if(x[1]<'1' || x[1]>'8'){throw logic_error(E);}
  if(x[2]<'a' || x[2]>'h'){throw logic_error(E);}
  if(x[3]<'1' || x[3]>'8'){throw logic_error(E);}
  
  char promo{' '};
  if(x.size()==5){promo=x[4];}
  int to=(x[2]-'a')+(x[3]-'1')*8;
  int from=(x[0]-'a')+(x[1]-'1')*8;
  Move m;
  
  if(x=="e1g1" && get_piece_from(e1).white_king() && get_piece_from(g1).no_piece() && get_piece_from(h1).white_rook()){
    m=convert_to_move(string("e1h1"));
  }else if(x=="e1c1" && get_piece_from(e1).white_king() && get_piece_from(c1).no_piece() && get_piece_from(a1).white_rook()){
    m=convert_to_move(string("e1a1"));
  }else if(x=="e8g8" && get_piece_from(e8).black_king() && get_piece_from(g8).no_piece() && get_piece_from(h8).black_rook()){
    m=convert_to_move(string("e8h8"));
  }else if(x=="e8c8" && get_piece_from(e8).black_king() && get_piece_from(c8).no_piece() && get_piece_from(a8).black_rook()){
    m=convert_to_move(string("e8a8"));
  }else if(promo=='q'){
    m=convert_to_move(Square(from),Square(to),promo_queen);
  }else if(promo=='r'){
    m=convert_to_move(Square(from),Square(to),promo_rook);
  }else if(promo=='b'){
    m=convert_to_move(Square(from),Square(to),promo_bishop);
  }else if(promo=='n'){
    m=convert_to_move(Square(from),Square(to),promo_knight);
  }else{
    m=convert_to_move(Square(from),Square(to));
  }

  make_move(m);
  look_for_check();
  m.check(check());
  unmake_move();

  return m;
}

Move ChessPosition::convert_to_move(Square square1, Square square2,move_t code)const
{
  Move m;
  m.from(square1);
  m.to(square2);
  m.code(code);
  m.piece(get_piece_from(m.from()));
  m.capture(get_piece_from(m.to()));
  m.xcode(xnormal_move);
  
  if(m.piece().no_piece())
    {
      m.code(nullmove);
    }
  else if(m.piece().white_king() && m.capture().white_rook())
    {
      m.code(castling);
      m.capture(NoPiece());
    }
  else if(m.piece().black_king() && m.capture().black_rook())
    {
      m.code(castling);
      m.capture(NoPiece());
    }
  else if(m.piece().white_pawn())
    {
      if((m.to()-m.from()==7 || m.to()-m.from()==9) && m.capture().no_piece() && get_piece_from(m.to()-8).black_pawn()){
	m.code(en_passant);
	m.capture(BlackPawn());
      }
      if(m.code()==0 && m.from()>=a7 && m.from()<=h7){m.code(code);}
    }
  else if(m.piece().black_pawn())
    {
      if((m.to()-m.from()==-7 || m.to()-m.from()==-9) && m.capture().no_piece() && get_piece_from(m.to()+8).white_pawn()){
	m.code(en_passant);
	m.capture(WhitePawn());
      }
      if(m.code()==0 && m.from()>=a2 && m.from()<=h2){m.code(code);}
    }
  
  if(m.code()==castling){m.capture(NoPiece());}
  return m;
}


void ChessPosition::make_move(const Square& square1, const Square& square2, move_t c)
{
  Move m;
  m=convert_to_move(square1,square2,c);
  make_move(m);
  look_for_check();
  m.check(check());
  unmake_move();
  make_move(m);
}

void ChessPosition::make_move(const Move& m)
{
  m_make_move_counter++;

  
  //------------------------------ move on bitboard ---------------  
  //
  bp.make_move(m);
  //
  //------------------------------ move on bitboard ---------------  

  
  //------------------------------ castling rights -----------------
  //
  uint16_t castlingrights=m_hist.back().castlingrights;

  if(castlingrights!=0x0000)
    {
      if(m.piece().white_king()){castlingrights&=0xff00;}
      if(m.piece().black_king()){castlingrights&=0x00ff;}

      int sq{m.from().id()};
      if(sq<8){castlingrights&=~(1<<sq);}
      else if(sq>55 && sq<64){sq&=0x07;castlingrights&=~((1<<sq)<<8);}

      sq=m.to().id();
      if(sq<8){castlingrights&=~(1<<sq);}
      else if(sq>55 && sq<64){sq&=0x07;castlingrights&=~((1<<sq)<<8);}
    }

  //
  //------------------------------ castling rights -----------------
  

  // half moves counter:
  uint16_t half_moves;
  if(m.piece().pawn() || m.capture()){half_moves=0;}
  else{half_moves=m_hist.back().half_moves+1;}
  // EP square:
  Square ep{no_square};
  if(m.piece().pawn() && (m.from()-m.to()==16 || m.to()-m.from()==16)){
    if((m.piece().white_pawn() && (m.to().bitboard()<<1 | m.to().bitboard()>>1) & black_pawn()) ||
       (m.piece().black_pawn() && (m.to().bitboard()<<1 | m.to().bitboard()>>1) & white_pawn()) ){
      ep=Square((m.from().id()+m.to().id())>>1);
    }
  }
  
  //
  m_hist.push_back(m_hist.back());
  m_hist.back().castlingrights=castlingrights;
  m_hist.back().half_moves=half_moves;
  m_hist.back().last_move=m;
  m_hist.back().ep=ep;
  m_hist.back().make_move_counter=m_make_move_counter;
  //

  // pinned and discovering pieces:
  find_pinned_pieces();
  //

  // total movecounter:
  if(bp.white_to_move()){m_total_moves++;}

  // check status, hash, threefold repetition:
  if(m.check()){m_hist.back().check=true;}else{m_hist.back().check=false;}
  m_hh.make_move(m);
  m_rtf.add(m,m_hh.hash());
  
  // material index:
  m_materialindex.make_move(m);
  //
}



void ChessPosition::unmake_move()
{
  Move m;
  m=m_hist.back().last_move;

  //------------------------------ unmake_move on bit board ---------------  
  //
  bp.unmake_move(m);
  //
  //------------------------------ unmake_move on bit board ---------------  

  m_hist.pop_back();
  if(!bp.white_to_move()){m_total_moves--;}
  m_hh.unmake_move();
  m_rtf.remove();

  // material index:
  m_materialindex.unmake_move(m);
  //
}

Piece ChessPosition::get_piece_from(const Square& sq)const
{
  return Piece(bp[sq.id()]);
}

int ChessPosition::get_piece_id_from(const Square& sq)const
{
  return bp[sq.id()];
}

int ChessPosition::get_piece_id_from(const BitBoard& bb_sq)const
{
  int i=bb_sq.get_unique_square().id();
  return bp[i];
}

void ChessPosition::set_piece_on(const Square& sq,Piece p)
{
  bp.set_piece_on(sq,p.id());
  bp.compute_pieces();
}

void ChessPosition::remove_piece_from(const Square& sq,Piece p)
{
  bp.remove_piece_from(sq,p.id());
  bp.compute_pieces();
}

void ChessPosition::set_castlingrights(short x)
{
  m_hist.back().castlingrights=x;
}

int ChessPosition::stm()const
{
  return bp.stm();
}

void ChessPosition::switch_color_to_move()
{
  bp.switch_color_to_move();
}

unsigned short ChessPosition::castlingrights()const
{
  return m_hist.back().castlingrights;
}

Square ChessPosition::ep()const
{
  return m_hist.back().ep;
}

void ChessPosition::compute_materialindex()const
{
  MaterialIndex r;
  r.set_whitequeen(white_queen().popcount());
  r.set_whiterook(white_rook().popcount());
  r.set_whitebishop(white_bishop().popcount());
  r.set_whiteknight(white_knight().popcount());
  r.set_whitepawn(white_pawn().popcount());
  r.set_blackqueen(black_queen().popcount());
  r.set_blackrook(black_rook().popcount());
  r.set_blackbishop(black_bishop().popcount());
  r.set_blackknight(black_knight().popcount());
  r.set_blackpawn(black_pawn().popcount());
  m_materialindex=r;
}

MaterialIndex ChessPosition::materialindex()const
{
  return m_materialindex;
}

uint64_t ChessPosition::get_hash()const
{
  return m_hh.hash();
}

uint64_t ChessPosition::get_pawnhash()const
{
  return m_hh.pawnhash();
}

int ChessPosition::get_dfz()const
{
  return m_hist.back().half_moves;
}

int ChessPosition::nobelmen()const
{
  int r=(white_knight()|black_knight()).popcount()*knight_value;
  r+=(white_bishop()|black_bishop()).popcount()*bishop_value;
  r+=(white_rook()|black_rook()).popcount()*rook_value;
  r+=(white_queen()|black_queen()).popcount()*queen_value;
  return r;
}

Square ChessPosition::white_kingsfield()const
{
  return m_hist.back().white_kingsfield;
}

Square ChessPosition::black_kingsfield()const
{
  return m_hist.back().black_kingsfield;
}

string ChessPosition::fen()const
{
  string r;
  char counter='0';
  for(int i=0;i<=56;i+=8)
    {
      for(int square=a8-i;square<=h8-i;square++)
	{
	  Piece p{get_piece_from(square)};
	  if(p.no_piece()){counter++;}
	  else
	    {
	      if(counter!='0'){r+=counter;counter='0';}
	      if(p.white_piece())
		{
		  if(p.white_pawn()){r+='P';}
		  else if(p.white_knight()){r+='N';}
		  else if(p.white_bishop()){r+='B';}
		  else if(p.white_rook()){r+='R';}
		  else if(p.white_queen()){r+='Q';}
		  else if(p.white_king()){r+='K';}
		}
	      else if(p.black_piece())
		{
		  if(p.black_pawn()){r+='p';}
		  else if(p.black_knight()){r+='n';}
		  else if(p.black_bishop()){r+='b';}
		  else if(p.black_rook()){r+='r';}
		  else if(p.black_queen()){r+='q';}
		  else if(p.black_king()){r+='k';}
		}
	    }
	}
      if(counter!='0'){r+=counter;counter='0';}
      if(i<56){r+="/";}
    }
  if(bp.white_to_move()){r+=" w ";}else{r+=" b ";}
  if((m_hist.back().castlingrights&0x7e7e)==0 && !m_chess960){
    if(m_hist.back().castlingrights&0x0080){r+="K";}
    if(m_hist.back().castlingrights&0x0001){r+="Q";}
    if(m_hist.back().castlingrights&0x8000){r+="k";}
    if(m_hist.back().castlingrights&0x0100){r+="q";}
  }
  else{
    if(m_hist.back().castlingrights&0x0001){r+="A";}
    if(m_hist.back().castlingrights&0x0002){r+="B";}
    if(m_hist.back().castlingrights&0x0004){r+="C";}
    if(m_hist.back().castlingrights&0x0008){r+="D";}
    if(m_hist.back().castlingrights&0x0010){r+="E";}
    if(m_hist.back().castlingrights&0x0020){r+="F";}
    if(m_hist.back().castlingrights&0x0040){r+="G";}
    if(m_hist.back().castlingrights&0x0080){r+="H";}
    if(m_hist.back().castlingrights&0x0100){r+="a";}
    if(m_hist.back().castlingrights&0x0200){r+="b";}
    if(m_hist.back().castlingrights&0x0400){r+="c";}
    if(m_hist.back().castlingrights&0x0800){r+="d";}
    if(m_hist.back().castlingrights&0x1000){r+="e";}
    if(m_hist.back().castlingrights&0x2000){r+="f";}
    if(m_hist.back().castlingrights&0x4000){r+="g";}
    if(m_hist.back().castlingrights&0x8000){r+="h";}
  }
  if(m_hist.back().castlingrights==0){r+="-";}
  if(m_hist.back().ep==no_square){r+=" - ";}
  else
    {
      string m{" "};m+=m_hist.back().ep.name();m+=" ";r+=m;
    }
  
  {
    ostringstream ostr;
    ostr << m_hist.back().half_moves << " ";
    r+=ostr.str();
  }

  {
    ostringstream ostr;
    ostr << m_total_moves;
    r+=ostr.str();
  }
  return r;
}

string ChessPosition::epd()const
{
  string r,t;
  string s=fen();
  istringstream iss(s);
  for(int i=0;i<4;i++){    
    iss >> t;
    r+=t+" ";
  }
  r.pop_back();
  return r;
}

string ChessPosition::initial_position()const
{
  return m_initial_position;
}

void ChessPosition::fen(string s)
{
  clear();
  Piece p;
  string s1,s2,s3,s4,s5{"0"},s6{"1"};
  istringstream istr(s);
  istr >> s1 >> s2 >> s3 >> s4 >> s5 >> s6;
  try{stoi(s5);}catch(...){s5="0";}
  try{stoi(s6);}catch(...){s6="1";}
  m_initial_position=s1+" "+s2+" "+s3+" "+s4+" "+s5+" "+s6;

  bp.fen(s);

  uint16_t castlingrights{0};
  uint16_t file;
  for(auto& c : s3)
    {
      switch(c){
      case 'K':
	file=0x0080;
	for(Square s=h1;s>=c1;s--){
	  if(get_piece_from(s).white_rook()){
	    break;
	  }
	  file=file>>1;
	}
	castlingrights |= file;
	break;
      case 'Q':
	file=0x0001;
	for(Square s=a1;s<=f1;s++){
	  if(get_piece_from(s).white_rook()){
	    break;
	  }
	  file=file<<1;
	}
	castlingrights |= file;
	break;
      case 'k':
	file=0x8000;
	for(Square s=h8;s>=c8;s--){
	  if(get_piece_from(s).black_rook()){
	    break;
	  }
	  file=file>>1;
	}
	castlingrights |= file;
	break;
      case 'q':
	file=0x0100;
	for(Square s=a8;s<=f8;s++){
	  if(get_piece_from(s).black_rook()){
	    break;
	  }
	  file=file<<1;
	}
	castlingrights |= file;
	break;
      case 'A':castlingrights |= 0x0001;break;
      case 'B':castlingrights |= 0x0002;break;
      case 'C':castlingrights |= 0x0004;break;
      case 'D':castlingrights |= 0x0008;break;
      case 'E':castlingrights |= 0x0010;break;
      case 'F':castlingrights |= 0x0020;break;
      case 'G':castlingrights |= 0x0040;break;
      case 'H':castlingrights |= 0x0080;break;
      case 'a':castlingrights |= 0x0100;break;
      case 'b':castlingrights |= 0x0200;break;
      case 'c':castlingrights |= 0x0400;break;
      case 'd':castlingrights |= 0x0800;break;
      case 'e':castlingrights |= 0x1000;break;
      case 'f':castlingrights |= 0x2000;break;
      case 'g':castlingrights |= 0x4000;break;
      case 'h':castlingrights |= 0x8000;break;
      }
    }
  
  Square ep;
  switch(s4[0])
    {
    case 'a':if(s4[1]=='3'){ep=a3;}else{ep=a6;}break;
    case 'b':if(s4[1]=='3'){ep=b3;}else{ep=b6;}break;
    case 'c':if(s4[1]=='3'){ep=c3;}else{ep=c6;}break;
    case 'd':if(s4[1]=='3'){ep=d3;}else{ep=d6;}break;
    case 'e':if(s4[1]=='3'){ep=e3;}else{ep=e6;}break;
    case 'f':if(s4[1]=='3'){ep=f3;}else{ep=f6;}break;
    case 'g':if(s4[1]=='3'){ep=g3;}else{ep=g6;}break;
    case 'h':if(s4[1]=='3'){ep=h3;}else{ep=h6;}break;
    }
  
  uint16_t halfmoves{0};
  istringstream iss(s5);
  iss >> halfmoves;

  {
    int totalmoves{1};
    istringstream iss(s6);
    iss >> totalmoves;
    m_total_moves=totalmoves;
  }

  
  m_hist.back()={BitBoard(),BitBoard(),BitBoard(),BitBoard(),m_make_move_counter,castlingrights,halfmoves,NullMove(),ep,no_square,no_square,false};
  if(bp.white_king() && bp.black_king()){find_pinned_pieces();}
  m_hh.init(*this);
  m_rtf.init(m_hh.hash());
  look_for_check();
  compute_materialindex();
}

void ChessPosition::flip()
{
  bp.flip();
  uint16_t cr=m_hist.back().castlingrights;
  cr=cr<<8|cr>>8;
  m_hist.back().castlingrights=cr;
  if(m_hist.back().ep!=no_square){
    m_hist.back().ep=m_hist.back().ep.id()^56;
  }
  fen(fen());
}

void ChessPosition::flop()
{
  bp.flop();
  uint16_t cr2{0};
  uint16_t cr=m_hist.back().castlingrights;
  cr=cr<<8|cr>>8;
  for(int i=0;i<16;i++){
    cr2<<=1;
    cr2|=cr&0x1;
    cr>>=1;
  }
  m_hist.back().castlingrights=cr2;
  if(m_hist.back().ep!=no_square){
    int rank=m_hist.back().ep.rank_number();
    int file=m_hist.back().ep.file_number();
    m_hist.back().ep=8*rank+(7-file);
  }
  fen(fen());
}

const vector<history_t>& ChessPosition::history()const
{
  return m_hist;
}

vector<Move> ChessPosition::moves_history()const
{
  vector<Move> r;
  for(size_t i=1;i<m_hist.size();i++){
    r.push_back(m_hist[i].last_move);
  }
  return r;
}

string ChessPosition::history_str()const
{
  ostringstream oss;
  for(auto& m : moves_history()){oss << m.algebraic() << " ";}
  string r=oss.str();
  if(!r.empty()){
    r.pop_back();
  }
  return r;
}

const BitBoard& ChessPosition::white_pawn()const{return bp.white_pawn();}
const BitBoard& ChessPosition::white_knight()const{return bp.white_knight();}
const BitBoard& ChessPosition::white_bishop()const{return bp.white_bishop();}
const BitBoard& ChessPosition::white_rook()const{return bp.white_rook();}
const BitBoard& ChessPosition::white_queen()const{return bp.white_queen();}
const BitBoard& ChessPosition::white_king()const{return bp.white_king();}
const BitBoard& ChessPosition::black_pawn()const{return bp.black_pawn();}
const BitBoard& ChessPosition::black_knight()const{return bp.black_knight();}
const BitBoard& ChessPosition::black_bishop()const{return bp.black_bishop();}
const BitBoard& ChessPosition::black_rook()const{return bp.black_rook();}
const BitBoard& ChessPosition::black_queen()const{return bp.black_queen();}
const BitBoard& ChessPosition::black_king()const{return bp.black_king();}
const uint64_t& ChessPosition::piece(int x)const{return bp.piece(x);}
const BitBoard& ChessPosition::white_pieces()const{return bp.white_pieces();}
const BitBoard& ChessPosition::black_pieces()const{return bp.black_pieces();}
const BitBoard& ChessPosition::pieces()const{return bp.pieces();}

Square ChessPosition::white_weakest_attacker(const Square& s1)const
{
  BitBoard b,attackbb_N,attackbb_B,attackbb_R,attackbb_Q,bb_s1;
  Square s2,s3;
  bb_s1=s1.bitboard();

  b=bb_s1>>7 & not_a_file;
  b|=bb_s1>>9 & not_h_file;
  b&=white_pawn();
  if(b){return b.get_first_square();}

  b=m_atta.knight(bb_s1);
  b&=white_knight();
  if(b){return b.get_first_square();}

  attackbb_B=m_atta.bishop(bb_s1,pieces());
  b=white_bishop() & attackbb_B;
  if(b){return b.get_first_square();}

  attackbb_R=m_atta.rook(bb_s1,pieces());
  b=white_rook() & attackbb_R;
  if(b){return b.get_first_square();}

  attackbb_Q=attackbb_B | attackbb_R;
  b=white_queen() & attackbb_Q;
  if(b){return b.get_first_square();}

  b=m_atta.king(bb_s1);
  b&=white_king();
  if(b){return b.get_first_square();}

  return s3;
}

Square ChessPosition::black_weakest_attacker(const Square& s1)const
{
  BitBoard b,attackbb_B,attackbb_R,attackbb_Q,bb_s1;
  Square s2,s3;
  bb_s1=s1.bitboard();

  b=s1.bitboard()<<7 & not_h_file;
  b|=s1.bitboard()<<9 & not_a_file;
  b&=black_pawn();
  if(b){s2=b.get_first_square();return s2;}

  b=m_atta.knight(bb_s1);
  b&=black_knight();
  if(b){return b.get_first_square();}

  attackbb_B=m_atta.bishop(bb_s1,pieces());
  b=black_bishop() & attackbb_B;
  if(b){return b.get_first_square();}

  attackbb_R=m_atta.rook(bb_s1,pieces());
  b=black_rook() & attackbb_R;
  if(b){return b.get_first_square();}

  attackbb_Q=attackbb_B | attackbb_R;
  b=black_queen() & attackbb_Q;
  if(b){return b.get_first_square();}

  b=m_atta.king(bb_s1);
  b&=black_king();
  if(b){return b.get_first_square();}

  return s3;
}

BitBoard ChessPosition::white_attackers(const BitBoard& bb_s1)const
{
  BitBoard r,b,attackbb_B,attackbb_R,attackbb_Q;
  Square s2,s3;

  if(!bb_s1){return r;}
  b=bb_s1>>7 & not_a_file;
  b|=bb_s1>>9 & not_h_file;
  b&=white_pawn();
  r=b;

  b=m_atta.knight(bb_s1);
  b&=white_knight();
  r|=b;

  attackbb_B=m_atta.bishop(bb_s1,pieces());
  b=white_bishop() & attackbb_B;
  r|=b;

  attackbb_R=m_atta.rook(bb_s1,pieces());
  b=white_rook() & attackbb_R;
  r|=b;

  attackbb_Q=attackbb_B | attackbb_R;
  b=white_queen() & attackbb_Q;
  r|=b;

  b=m_atta.king(bb_s1);
  b&=white_king();
  r|=b;

  return r;
}

BitBoard ChessPosition::black_attackers(const BitBoard& bb_s1)const
{
  BitBoard r,b,attackbb_B,attackbb_R,attackbb_Q;
  Square s2,s3;
  
  if(!bb_s1){return r;}
  b=bb_s1<<7 & not_h_file;
  b|=bb_s1<<9 & not_a_file;
  b&=black_pawn();
  r=b;

  b=m_atta.knight(bb_s1);
  b&=black_knight();
  r|=b;

  attackbb_B=m_atta.bishop(bb_s1,pieces());
  b=black_bishop() & attackbb_B;
  r|=b;

  attackbb_R=m_atta.rook(bb_s1,pieces());
  b=black_rook() & attackbb_R;
  r|=b;

  attackbb_Q=attackbb_B | attackbb_R;
  b=black_queen() & attackbb_Q;
  r|=b;

  b=m_atta.king(bb_s1);
  b&=black_king();
  r|=b;

  return r;
}

BitBoard ChessPosition::white_blocked_attackers(const BitBoard& bb_s1)const
{
  BitBoard r,b,empty,attackbb_B,attackbb_R,attackbb_Q;
  Square s2,s3;

  b=bb_s1>>7 & not_a_file;
  b|=bb_s1>>9 & not_h_file;
  b&=white_pawn();
  r=b;

  b=m_atta.knight(bb_s1);
  b&=white_knight();
  r|=b;

  attackbb_B=m_atta.bishop(bb_s1,empty);
  b=white_bishop() & attackbb_B;
  r|=b;

  attackbb_R=m_atta.rook(bb_s1,empty);
  b=white_rook() & attackbb_R;
  r|=b;

  attackbb_Q=attackbb_B | attackbb_R;
  b=white_queen() & attackbb_Q;
  r|=b;

  b=m_atta.king(bb_s1);
  b&=white_king();
  r|=b;

  return r;
}

BitBoard ChessPosition::black_blocked_attackers(const BitBoard& bb_s1)const
{
  BitBoard r,b,empty,attackbb_B,attackbb_R,attackbb_Q;
  Square s2,s3;

  b=bb_s1<<7 & not_h_file;
  b|=bb_s1<<9 & not_a_file;
  b&=black_pawn();
  r|=b;

  b=m_atta.knight(bb_s1);
  b&=black_knight();
  r|=b;

  attackbb_B=m_atta.bishop(bb_s1,empty);
  b=black_bishop() & attackbb_B;
  r|=b;

  attackbb_R=m_atta.rook(bb_s1,empty);
  b=black_rook() & attackbb_R;
  r|=b;

  attackbb_Q=attackbb_B | attackbb_R;
  b=black_queen() & attackbb_Q;
  r|=b;

  b=m_atta.king(bb_s1);
  b&=black_king();
  r|=b;

  return r;
}

bool ChessPosition::white_attacks(const BitBoard& bb_s1)const
{
  if(white_attackers(bb_s1)){return true;}else{return false;}
}

bool ChessPosition::black_attacks(const BitBoard& bb_s1)const
{
  if(black_attackers(bb_s1)){return true;}else{return false;}
}

void ChessPosition::find_pinned_pieces()
{
  BitBoard segment;
  m_hist.back().pinned_pieces.clear();
  m_hist.back().discovering_pieces.clear();
  
  m_hist.back().black_kingsfield=bp.black_king().get_unique_square();
  BitBoard pinners{
    ((white_queen() | white_rook()) & m_con.rook_attacks(m_hist.back().black_kingsfield))
      | ((white_queen() | white_bishop()) & m_con.bishop_attacks(m_hist.back().black_kingsfield))};
  while(pinners){
    Square s=pinners.get_reset_square().get_unique_square();
    segment=m_con.line_segment(s,m_hist.back().black_kingsfield);
    if((pieces()&segment).popcount()==1){
      if(white_pieces()&segment){m_hist.back().discovering_pieces|=white_pieces()&segment;}
      if(black_pieces()&segment){m_hist.back().pinned_pieces|=black_pieces()&segment;}
    }
  }
  
  m_hist.back().white_kingsfield=bp.white_king().get_unique_square();
  pinners=((black_queen() | black_rook()) & m_con.rook_attacks(m_hist.back().white_kingsfield))
    | ((black_queen() | black_bishop()) & m_con.bishop_attacks(m_hist.back().white_kingsfield));
  while(pinners){
    Square s=pinners.get_reset_square().get_unique_square();
    segment=m_con.line_segment(s,m_hist.back().white_kingsfield);
    if((pieces()&segment).popcount()==1){
      if(black_pieces()&segment){m_hist.back().discovering_pieces|=black_pieces()&segment;}
      if(white_pieces()&segment){m_hist.back().pinned_pieces|=white_pieces()&segment;}
    }
  }
  
  m_hist.back().white_queencheck_squares=m_atta.queen(m_hist.back().black_kingsfield,pieces());
  m_hist.back().black_queencheck_squares=m_atta.queen(m_hist.back().white_kingsfield,pieces());

  return;
}

const BitBoard& ChessPosition::pinned_pieces()const
{
  return m_hist.back().pinned_pieces;
}

const BitBoard& ChessPosition::discovering_pieces()const
{
  return m_hist.back().discovering_pieces;
}

const BitBoard& ChessPosition::white_queencheck_squares()const
{
  return m_hist.back().white_queencheck_squares;
}

const BitBoard& ChessPosition::black_queencheck_squares()const
{
  return m_hist.back().black_queencheck_squares;
}

void ChessPosition::look_for_check()
{
  m_hist.back().check=false;
  if(bp.white_to_move()){
    if(black_attackers(white_king())){
      m_hist.back().check=true;
    }
  }
  else{
    if(white_attackers(black_king())){
      m_hist.back().check=true;
    }
  }
}

bool ChessPosition::check()const
{
  return m_hist.back().check;
}

const Move& ChessPosition::last_move()const
{
  return m_hist.back().last_move;
}

uint64_t ChessPosition::get_make_move_counter()const
{
  return m_make_move_counter;
}

bool ChessPosition::draw_by_threefold_repetition()const
{
  return m_rtf.repetition();
}

bool ChessPosition::draw_by_fifty_move_rule()const
{
  if(m_hist.back().half_moves>=100){return true;}
  else{return false;}
}

bool ChessPosition::operator==(const ChessPosition& x)const
{
  if(bp!=x.bp){return false;}
  if(m_total_moves!=x.m_total_moves)return false;
  if(m_hist.back().half_moves!=x.m_hist.back().half_moves)return false;
  if(m_hist.back().castlingrights!=x.m_hist.back().castlingrights)return false;
  if(m_hist.back().ep!=x.m_hist.back().ep)return false;
  if(m_hist.back().pinned_pieces!=x.m_hist.back().pinned_pieces)return false;
  if(m_hist.back().discovering_pieces!=x.m_hist.back().discovering_pieces)return false;
  return true;
}

bool ChessPosition::operator!=(const ChessPosition& x)const
{
  
  return !(*this==x);
}

const int& ChessPosition::operator[](const int& i)const
{
  return bp[i];
}

void ChessPosition::force_move(const Move& m)
{
  bp.make_move(m);
}

void ChessPosition::undo_forced_move(const Move& m)
{
  bp.unmake_move(m);
}

int ChessPosition::id()const
{
  return m_cp_id;
}

string ChessPosition::dump()const
{
  ostringstream oss;
  oss << bp.dump() << endl;
  oss << "total moves: " << m_total_moves << endl;
  oss << "hash: " << hex << get_hash() << dec << endl;
  oss << boolalpha;
  oss << "check: " << m_hist.back().check << endl;
  oss << "pinned pieces: " << m_hist.back().pinned_pieces.dump() << endl;
  oss << "discovering pieces: " << m_hist.back().discovering_pieces.dump() << endl;
  return oss.str();
}

void ChessPosition::set_logging(bool x)
{
  if(x && !m_log){
    ostringstream oss;
    oss << "cposition_" << m_cp_id << ".log";
    m_ofs.open(oss.str());
    if(!m_ofs.good()){std::cerr << "ChessPosition open file error" << endl;}
  }
  else if(m_log && !x){
    m_ofs.close();
  }
  m_log=x;
}

void ChessPosition::set_chess960(bool x)
{
  m_chess960=x;
}
