/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifndef POSITION_H
#define POSITION_H

#include"../eval/materialindex.h"
#include"bposition.h"
#include"../common/piece.h"
#include"../common/connections.h"
#include"../common/bitboard.h"
#include"../common/tables.h"
#include"../board/hasher.h"
#include"../board/rep3filter.h"
#include"../common/attacks.h"
#include"../movegen/move.h"
#include"../search/timer.h"
#include<vector>
#include<string>
#include<fstream>
using std::ofstream;
using std::string;
using std::vector;

class Square;
class MoveList;
class MoveGenerator;
class MaterialIndex;

struct history_t
{
  BitBoard pinned_pieces;
  BitBoard discovering_pieces;
  BitBoard white_queencheck_squares;
  BitBoard black_queencheck_squares;
  uint64_t make_move_counter;
  uint16_t castlingrights;
  uint16_t half_moves;
  Move last_move;
  Square ep;
  Square white_kingsfield;
  Square black_kingsfield;
  bool check;
};

class ChessPosition
{
public:
  //----------------constructor/destructor/assignment--------------------
  //
  ChessPosition(const Tables& x);
  ChessPosition(const ChessPosition&);
  ChessPosition& operator=(const ChessPosition&);
  ~ChessPosition();
  //---------------------------------------------------------------------


  //----------------changing the board-----------------------------------
  //
  //-------- set position:
  void clear();
  void newgame();
  void newgame960(int=-1);
  void set_piece_on(const Square&,Piece);
  void remove_piece_from(const Square&,Piece);
  void set_castlingrights(short);
  void fen(string);
  void flip();
  void flop();
  
  //-------- make moves:
  void make_move(const Square&,const Square&,move_t=normal_move);
  void make_move(const Move&);
  void make_quies_move(const Move&);
  void unmake_move();
  void unmake_quies_move(const Move&);
  void force_move(const Move&);
  void undo_forced_move(const Move&);
  void switch_color_to_move();
  //---------------------------------------------------------------------


  //--------------------------queries------------------------------------
  //
  int stm()const;
  Piece get_piece_from(const Square&)const;
  int get_piece_id_from(const Square&)const;
  int get_piece_id_from(const BitBoard&)const;
  bool white_attacks(const BitBoard&)const;
  bool black_attacks(const BitBoard&)const;

  bool check()const;
  bool draw_by_threefold_repetition()const;
  bool draw_by_fifty_move_rule()const;

  unsigned short castlingrights()const;
  Square ep()const;
  int nobelmen()const;
  const Move& last_move()const;
  const vector<history_t>& history()const;
  vector<Move> moves_history()const;
  string history_str()const;
  string fen()const;
  string epd()const;
  string initial_position()const;
  MaterialIndex materialindex()const;
  uint64_t get_hash()const;
  uint64_t get_pawnhash()const;
  int get_dfz()const;
  Square white_kingsfield()const;
  Square black_kingsfield()const;
  uint64_t get_make_move_counter()const;

  bool operator==(const ChessPosition&)const;
  bool operator!=(const ChessPosition&)const;
  const int& operator[](const int&)const;
  int id()const;
  string dump()const;
  //
  
  //--------------------------general functions--------------------------
  //
  Move convert_to_move(Square, Square,move_t=normal_move)const;
  Move convert_to_move(const string&);
  void compute_materialindex()const;
  void set_logging(bool);
  void set_chess960(bool);
  //

  //--------------------------piece positions: bitboards-----------------
  //
  const BitBoard& white_pawn()const;
  const BitBoard& white_knight()const;
  const BitBoard& white_bishop()const;
  const BitBoard& white_rook()const;
  const BitBoard& white_queen()const;
  const BitBoard& white_king()const;
  const BitBoard& black_pawn()const;
  const BitBoard& black_knight()const;
  const BitBoard& black_bishop()const;
  const BitBoard& black_rook()const;
  const BitBoard& black_queen()const;
  const BitBoard& black_king()const;
  const uint64_t& piece(int)const;
  const BitBoard& white_pieces()const;
  const BitBoard& black_pieces()const;
  const BitBoard& pieces()const;
  Square white_weakest_attacker(const Square&)const;
  Square black_weakest_attacker(const Square&)const;
  BitBoard white_attackers(const BitBoard&)const;
  BitBoard black_attackers(const BitBoard&)const;
  BitBoard white_blocked_attackers(const BitBoard&)const;
  BitBoard black_blocked_attackers(const BitBoard&)const;
  //


  //----------------MoveGenerator functions------------------------------
  //
  void find_pinned_pieces();
  const BitBoard& pinned_pieces()const;
  const BitBoard& discovering_pieces()const;
  void look_for_check();
  //

  const BitBoard& white_queencheck_squares()const;
  const BitBoard& black_queencheck_squares()const;
  

  
private:
  static int cp_id;
  int m_cp_id;
  BoardPosition bp;
  const Tables& m_tables;
  const Attacks& m_atta;
  const Connections& m_con;
  Hasher<1> m_hh;
  Rep3filter m_rtf;

  uint16_t m_total_moves;
  vector<history_t> m_hist;
  string m_initial_position;
  mutable MaterialIndex m_materialindex;
  bool m_chess960{false};

  uint64_t m_make_move_counter;

  bool m_log{false};
  ofstream m_ofs;
};

     

#endif
