/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include<iostream>
#include<sstream>
#include<iomanip>
#include<stdexcept>
#include<cassert>
#include"bitboard.h"
using std::logic_error;
using std::cout;
using std::endl;
using std::ostringstream;



const array<char,128> BitBoard::m_first_square_id=[]()->const array<char,128>
{
  array<char,128> r;
  uint64_t s;
  for(int i=0;i<64;i++){
    s=1ull<<i;
    s*=988808377918865949ull;
    s>>=57;
    r[s]=i;
  }
  r[0]=no_square;
  return r;
}();

const array<char,1<<16> BitBoard::m_popcnt16=[]()->const array<char,1<<16>
{
  array<char,1<<16> r;
  for(unsigned int i=0;i<1<<16;i++){
    char count{0};
    uint16_t x=i;
    while(x){
      ++count;
      x&=x-1;
    }
    r[i]=count;
  }
  return r;  
}();

BitBoard::BitBoard()
{
}

BitBoard::BitBoard(const BitBoard& x): bb{x.bb}
{
}

BitBoard::BitBoard(const uint64_t& x): bb{x}
{
}

BitBoard::BitBoard(initializer_list<Square> x)
{
  bb=0;
  for(auto& s : x){set_square(s);}
}

void BitBoard::clear()
{
  bb=0;
}

void BitBoard::set_square(const Square& x)
{
  uint64_t cc{1};
  cc<<=x.id();
  bb|=cc;
}

void BitBoard::reset_square(const Square& x)
{
  uint64_t cc{1};
  cc<<=x.id();
  bb&=~cc;
}

void BitBoard::flip()
{
  uint64_t r;
  r=bb<<56;
  r|=(bb&0x000000000000FF00)<<40;
  r|=(bb&0x0000000000FF0000)<<24;
  r|=(bb&0x00000000FF000000)<<8;
  r|=(bb&0x000000FF00000000)>>8;
  r|=(bb&0x0000FF0000000000)>>24;
  r|=(bb&0x00FF000000000000)>>40;
  r|=(bb&0xFF00000000000000)>>56;
  bb=r;
}

void BitBoard::flop()
{
  uint64_t r;
  r=(bb&0x0101010101010101)<<7;
  r|=(bb&0x0202020202020202)<<5;
  r|=(bb&0x0404040404040404)<<3;
  r|=(bb&0x0808080808080808)<<1;
  r|=(bb&0x1010101010101010)>>1;
  r|=(bb&0x2020202020202020)>>3;
  r|=(bb&0x4040404040404040)>>5;
  r|=(bb&0x8080808080808080)>>7;
  bb=r;
}

Square BitBoard::get_first_square()const
{
  uint64_t s=bb & ((bb-1)^bb);
  s=s*988808377918865949ull;
  s>>=57;
  return Square{m_first_square_id[s]};
}

Square BitBoard::get_unique_square()const
{
  uint64_t s=bb*988808377918865949ull;
  s>>=57;
  return Square{m_first_square_id[s]};
}

BitBoard BitBoard::get_reset_square()
{
  uint64_t cc{bb-1};
  cc^=bb;
  cc&=bb;
  bb^=cc;
  return BitBoard(cc);
}

int BitBoard::get_reset_square(uint64_t& x)
{
  uint64_t cc{x-1};
  cc^=x;
  cc&=x;
  x^=cc;
  uint64_t s=cc*988808377918865949ull;
  s>>=57;
  return m_first_square_id[s];
}

vector<Square> BitBoard::get_squares()const
{
  vector<Square> r;
  uint64_t cc=bb;

  while(cc)
    {
      uint64_t dd{cc-1};
      dd^=cc;
      dd&=cc;
      cc^=dd;
      dd=dd*988808377918865949ull;
      dd>>=57;
      r.push_back(Square{m_first_square_id[dd]});      
    }
  return r;
}

int BitBoard::popcount()const
{
  return m_popcnt16[bb & 0xFFFF] + m_popcnt16[bb>>16 & 0xFFFF] + m_popcnt16[bb>>32 & 0xFFFF] + m_popcnt16[bb>>48];
}

int BitBoard::popcount(uint64_t x)
{
  return m_popcnt16[x & 0xFFFF] + m_popcnt16[x>>16 & 0xFFFF] + m_popcnt16[x>>32 & 0xFFFF] + m_popcnt16[x>>48];
}
  
bool BitBoard::operator<(BitBoard x)const
{
  return bb<x.bb;
}
  
bool BitBoard::operator>(BitBoard x)const
{
  return bb>x.bb;
}
  
bool BitBoard::operator==(BitBoard x)const
{
  return bb==x.bb;
}
  
bool BitBoard::operator!=(BitBoard x)const
{
  return bb!=x.bb;
}
  
BitBoard::operator bool()const
{
  return bb!=0;
}

BitBoard& BitBoard::operator&=(const BitBoard& x)
{
  bb&=x.bb;
  return *this;
}

BitBoard& BitBoard::operator|=(const BitBoard& x)
{
  bb|=x.bb;
  return *this;
}

BitBoard& BitBoard::operator^=(const BitBoard& x)
{
  bb^=x.bb;
  return *this;
}

BitBoard& BitBoard::operator&=(const uint64_t x)
{
  bb&=x;
  return *this;
}

BitBoard& BitBoard::operator|=(const uint64_t x)
{
  bb|=x;
  return *this;
}

BitBoard& BitBoard::operator^=(const uint64_t x)
{
  bb^=x;
  return *this;
}

BitBoard& BitBoard::operator>>=(int x)
{
  bb>>=x;
  return *this;
}

BitBoard& BitBoard::operator<<=(int x)
{
  bb<<=x;
  return *this;
}

BitBoard BitBoard::operator~()const
{
  BitBoard x;
  x.bb=~bb;
  return x;
}

const uint64_t& BitBoard::to_ull(const BitBoard& x)
{
  return x.bb;
}

void BitBoard::write()const
{
  cout << endl;
  std::cout << std::hex << std::setw(16) << std::setfill('0') << bb << std::dec << std::endl;
  for(char i=0;i<8;i++)
    {
      for(char j=0;j<8;j++)
	{
	  uint64_t cc=1;
	  cc<<=(7-i)*8+j;
	  if(bb&cc){std::cout << '1';}else{std::cout << '0';}
	}
      std::cout << std::endl;
    }
}

string BitBoard::dump()const
{
  ostringstream oss;
  oss << std::hex << std::setw(16) << std::setfill('0') << bb << std::dec << std::endl;
  for(char i=0;i<8;i++)
    {
      for(char j=0;j<8;j++)
	{
	  uint64_t cc=1;
	  cc<<=(7-i)*8+j;
	  if(bb&cc){oss << '1';}else{oss << '0';}
	}
      oss << std::endl;
    }
  return oss.str();
}

BitBoard operator&(const BitBoard& x,const BitBoard& y)
{
  return BitBoard(x.bb & y.bb);
}

BitBoard operator|(const BitBoard& x,const BitBoard& y)
{
  return BitBoard(x.bb | y.bb);
}

BitBoard operator^(const BitBoard& x,const BitBoard& y)
{
  return BitBoard(x.bb ^ y.bb);
}

BitBoard operator<<(const BitBoard& x,int n)
{
  return BitBoard(x.bb << n);
}

BitBoard operator>>(const BitBoard& x,int n)
{
  return BitBoard(x.bb >> n);
}

