/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"search/analyze.h"
#include"engine.h"
#include"common/numerals.h"
#include<cmath>
#include<chrono>
#include<functional>
#include<deque>
#include<string>
#include<iostream>
#include<iomanip>
#include<fstream>
#include<sstream>
#include<stdexcept>
using std::istringstream;
using std::ostringstream;
using std::logic_error;
using std::deque;
using std::string;
using std::cout;
using std::endl;
using namespace std::chrono;


Engine::Engine(Communicator& cc, ofstream& ofs) : m_com(cc), m_hh(m_t), m_ofs(ofs)
{
  time_point<high_resolution_clock> t0,t1;
  t1=high_resolution_clock::now();
  auto d=t1-t0;
  m_id=d.count()&0x1ffull;

  book.set_name("/usr/local/games/galjoen.files/Galjoen.bin");
  try{book.open();}
  catch(logic_error& E)
    {
      book.set_name("/usr/games/galjoen.files/Galjoen.bin");
      try{book.open();}
      catch(logic_error& E)
	{
	  book.set_name("Galjoen.bin");
	  try{book.open();}
	  catch(logic_error& E)
	    {
	      m_own_book_available=false;
	    }
	}
    }
  m_status=e_idle;
  cp.newgame();
  m_ana.init(cp);
  thread tmp_thread{std::ref(m_ana)};
  m_t1=move(tmp_thread);
  
}

Engine::~Engine()
{
  m_t1.join();
}

void Engine::send_info(Analysis& a)
{
  string oline;
  int depth=a.get_iteration();
  uint64_t nodes=a.get_node_count();
  uint64_t time=a.get_running_time();
  if(time<1)return;
  uint64_t nps=nodes / time * 1000;
  for(int i=0;i<m_mpv.size();i++){
    auto iter=m_mpv.begin()+i;
    ostringstream ostr;
    ostr << "info depth " << depth;
    ostr << " seldepth " << a.get_seldepth();
    if(m_mpv.number()>1){
      ostr << " multipv " << i+1;
    }
    if(iter->score()<mate_value-max_search_depth && iter->score()>-mate_value+max_search_depth){
      ostr << " score cp " << iter->score();
    }else{
      int n;
      if(iter->score()>=0){
	n=(mate_value - iter->score())/2+1;
      }else{
	n=(-mate_value - iter->score())/2;
      }
      ostr << " score mate " << n;
    }
    ostr << " time " << time;
    ostr << " nodes " << nodes;
    ostr << " nps " << nps;
    ostr << " pv ";
    auto v=*iter;
    for(auto& m : v){ostr << m.algebraic(m_chess960) << " ";}
    oline=ostr.str();
    m_ofs << "engine -> " << oline << endl;
    m_com.to_interface(oline);
  }
}

void Engine::send_hashfull(Analysis& a)
{
  string oline;
  ostringstream ostr;
  int hashfull=a.get_hashfull();
  ostr << "info hashfull " << hashfull;
  oline+=ostr.str();
  m_ofs << "engine -> " << oline << endl;
  m_com.to_interface(oline);		    
}

void Engine::set_action(char x)
{
  switch(x)
    {
    case stop_thinking:
      m_ana.set_action(stop_thinking);
      while(m_ana.get_status()==thinking){}
      break;
    case start_iterations:
      m_ana.set_action(start_iterations);
      while(m_ana.get_status()!=thinking){}
      break;
    }
}

bool Engine::equal_options(string x,string y)
{
  for(auto& c : x){c=toupper(c);}
  for(auto& c : y){c=toupper(c);}
  return x==y;
}

void Engine::from_console()
{
    m_command=m_com.from_interface();
    while(m_command!=""){
      istringstream istr(m_command);
      string command;
      if(!istr.eof()){istr >> command;}
      if(command.size()>0){m_ofs << "engine <- " << m_command << endl;}
      if(command=="ucinewgame"){
	cp.newgame();
	m_ana.clear_eval();
	m_ana.clear_hash();
	m_ana.clear_killer();
	m_ana.clear_history();
	m_previous_position.clear();
      }else if(command=="eval"){
	cout << m_ana.static_eval_report();
      }else if(command=="bench"){
	int64_t nps,nodes,time,total_time{0},total_nodes{0},eval,depth{12};
	float ebf;
	vector<string> vs;
	vs.push_back("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1");
	vs.push_back("rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w - - 0 1");
	vs.push_back("r2q2k1/pp1rbppp/4pn2/2P5/1P3B2/6P1/P3QPBP/1R3RK1 w - - 0 1");     // WAC.029
	for(auto& position : vs){
	  cp.fen(position);
	  m_ana.init(cp);
	  m_ana.clear_eval();
	  m_ana.clear_hash();
	  m_ana.clear_killer();
	  m_ana.clear_history();
	  m_ana.reset();
	  m_ana.set_hash_size(128);
	  m_ana.set_wtime(6000000);
	  m_ana.set_btime(6000000);
	  m_ana.set_movestogo(1);
	  Timer tim;
	  m_ana.set_depth(depth);
	  cout << "position: " << cp.fen() << "\t" << std::flush;
	  tim.start();
	  m_ana.iterative_search();
	  tim.stop();
	  time=tim.total_time().milliseconds();
	  nodes=m_ana.get_node_count();
	  if(time > 0){nps=nodes*1000/time;}else{nps=0;}
	  ebf=m_ana.get_ebf();
	  eval=m_ana.eval();
	  string solution=m_ana.best_move().algebraic();
	  cout << "  depth=" << depth
	       << "  nps=" << nps 
	       << "  nodes=" << nodes 
	       << "  time=" << std::setprecision(4) << time/1000.0
	       << "  ebf=" << std::setprecision(2) << ebf
	       << "  eval=" << std::setprecision(2) << eval
	       << "  bestmove " << solution
	       << endl;
	  total_nodes+=nodes;
	  total_time+=time;
	}
	cout << "-------------------------------------" << endl;
	cout << "nodes: \t" << total_nodes << endl;
	cout << "time: \t" << total_time << endl;
	cout << "nps:  \t" << static_cast<int>(round(total_nodes*1000.0/total_time)) << endl;
	cout << "-------------------------------------" << endl;
      }else if(command=="usermove"){
	string s;
	istr >> s;
	Move m;
	try{
	  m=cp.convert_to_move(s);
	}
	catch(...){
	  std::cerr << cp.fen() << "    move=" << s << endl;
	  throw logic_error("Engine::from_console() error");
	}
	cp.make_move(m);
	m_ana.init(cp);
	m_ana.reset();
      }else if(command=="position"){
	if(m_ana.get_status()==thinking){set_action(stop_thinking);}

	while(!m_command.empty() && m_command.back()==' '){m_command.pop_back();}
	unsigned startpos_lastmove=m_command.size()-1;
	while(startpos_lastmove-1>0 && m_command[startpos_lastmove-1]!=' '){startpos_lastmove--;}
	string last_move;
	for(unsigned i=startpos_lastmove;i<m_command.size();i++){
	  last_move.push_back(m_command[i]);
	}
	Move m;
	if(m_previous_position==m_command.substr(0,startpos_lastmove-1)){
	  //---------------------------- make last move ------------------------
	  //
	  m_previous_position=m_command;
	  m=cp.convert_to_move(last_move);
	  cp.make_move(m);
	  m_ana.init(cp);
	  m_ana.reset();
	  //
	  //---------------------------- make last move ------------------------
	}else{
	  //---------------------------- setup position -----------------------
	  //
	  m_previous_position=m_command;
	  string s1,s2,s3,s4,s5,s6;
	  istr >> s1;
	  if(s1=="startpos"){
	    cp.newgame();
	  }else if(s1=="testpos"){
	    cp.fen("b4bN1/4p1p1/1Q1p2K1/q1np4/3p1Rnk/6pr/4Rppr/3N4 w - - 0 1");
	  }else if(s1=="fen"){
	    istr >> s1 >> s2 >> s3 >> s4 >> s5 >> s6;
	    cp.fen(s1+" "+s2+" "+s3+" "+s4+" "+s5+" "+s6);
	  }
	  istr >> s1;
	  if(s1=="moves"){
	    s1="";
	    while(!istr.eof()){
	      istr >> s1;
	      if(s1.size()>0){
		Move m=cp.convert_to_move(s1);
		cp.make_move(m);
		s1.clear();
	      }
	    }
	  }
	  m_ana.init(cp);
	  m_ana.reset();
	  //
	  //---------------------------- setup position -----------------------	  
	}
      }else if(command=="go"){
	m_ana.reset();
	m_ana.set_movestogo(40);
	m_ana.set_depth(max_search_depth);
	m_ana.reset_max_nodes();
	BKmove m;
	if(m_own_book && cp.history().size()<=own_book_maxdepth){
	  m_hh.init(cp);
	  uint64_t key=m_hh.hash();
	  m=book.get_move(key);
	}
	if(m_own_book && !m.nomove()){
	  Move mm=m(cp);
	  command="bestmove " + mm.algebraic(m_chess960);
	  m_com.to_interface(command);
	}else{
	  if(m_status==e_thinking){m_ofs << "Error: received go while already thinking" << endl;return;}
	  m_status=e_thinking;
	  string s1;
	  int wtime,btime,winc,binc,movestogo,depth,nodes,mate,movetime;
	  while(true){
	    s1.clear();
	    istr >> s1;
	    if(s1=="wtime"){
	      istr >> wtime;
	      m_ana.set_wtime(wtime);
	    }else if(s1=="btime"){
	      istr >> btime;
	      m_ana.set_btime(btime);
	    }else if(s1=="winc"){istr >> winc;m_ana.set_winc(winc);}
	    else if(s1=="binc"){istr >> binc;m_ana.set_binc(binc);}
	    else if(s1=="movestogo"){istr >> movestogo;m_ana.set_movestogo(movestogo);}
	    else if(s1=="depth"){istr >> depth;m_ana.set_depth(depth);}
	    else if(s1=="nodes"){istr >> nodes;m_ana.set_nodes(nodes);}
	    else if(s1=="mate"){istr >> mate;m_ana.set_mate(mate);}
	    else if(s1=="movetime"){istr >> movetime;m_ana.set_movetime(movetime);}
	    else if(s1=="infinite"){m_ana.set_infinite(true);m_status=e_infinite;}
	    else if(s1=="ponder"){m_ana.set_pondering(true);m_status=e_pondering;}
	    else if(s1=="searchmoves"){
	      cp=m_ana.chessposition();
	      vector<Move> vm;
	      while(!istr.eof()){
		s1="";
		istr >> s1;
		if(s1==""){break;}
		Move m=cp.convert_to_move(s1);
		vm.push_back(m);
	      }
	      m_ana.set_searchmoves(vm);
	    }  
	    else{break;}
	  }
	  m_ana.set_action(start_iterations);
	}
      }else if(command=="ponderhit"){
	m_ana.ponderhit();
	m_status=e_thinking;
      }else if(command=="setoption"){
	string s1;
	istr >> s1;
	if(s1=="name"){
	  istr >> s1;
	  if(equal_options(s1,"Hash")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> m_memory;
	      if(m_memory>0){
		m_ana.set_hash_size(m_memory);
	      }
	    }
	  }
	  if(equal_options(s1,"Threads")){
	    istr >> s1;
	    if(s1=="value"){
	      int max_threads;istr >> max_threads;
	      m_ana.set_max_threads(max_threads);
	    }
	  }
	  else if(equal_options(s1,"Ponder")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> s1;
	      if(s1=="true"){m_ponder=true;}else{m_ponder=false;}
	    }
	  }
	  else if(equal_options(s1,"OwnBook")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> s1;
	      if(s1=="true" && m_own_book_available){m_own_book=true;}else{m_own_book=false;}
	    }
	  }
	  else if(equal_options(s1,"UCI_LimitStrength")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> s1;
	      if(s1=="true"){
		m_ana.set_uci_limitstrength(true);
	      }else{
		m_ana.set_uci_limitstrength(false);
	      }
	    }
	  }
	  else if(equal_options(s1,"UCI_Elo")){
	    istr >> s1;
	    if(s1=="value"){
	      int elo;istr >> elo;
	      m_ana.set_uci_elo(elo);
	    }
	  }
	  else if(equal_options(s1,"UCI_Chess960")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> s1;
	      if(s1=="true"){m_chess960=true;}else{m_chess960=false;}
	    }
	  }
	  else if(equal_options(s1,"MultiPV")){
	    istr >> s1;
	    if(s1=="value"){
	      int mpv;istr >> mpv;
	      m_ana.set_multipv(mpv);
	    }
	  }
	  else if(equal_options(s1,"Debug")){
	    istr >> s1;
	    if(s1=="value"){
	      istr >> s1;
	      if(s1=="true"){
		m_debug=true;
		m_ana.set_log(true);
	      }else{
		m_debug=false;
		m_ana.set_log(false);
	      }
	      if(m_ofs && m_debug==false){m_ofs.close();}
	      else if(!m_ofs && m_debug==true){
		ostringstream oss;
		oss << "uci_eng_" << m_id << ".log";
		m_ofs.open(oss.str());
	      }
	    }
	  }
	}
      }else if(command=="isready"){
	m_ofs << "engine -> " << "readyok" << endl;
	m_com.to_interface("readyok");
      }else if(command=="stop" && (m_status==e_thinking || m_status==e_pondering || m_status==e_infinite)){
	set_action(stop_thinking);
	m_status=e_idle;
	//------------------------------   don't forget the best move: --------------------
	while(m_ana.get_status()==thinking){std::this_thread::sleep_for(std::chrono::milliseconds(5));}
	if(m_mpv.version()<m_ana.mpv().version()){
	  m_mpv=m_ana.mpv();
	  m_mpv.set_version(m_ana.mpv().version());
	  send_info(m_ana);
	}
	Move bestmove=m_ana.best_move();
	cp.make_move(bestmove);
	command="bestmove " + bestmove.algebraic(m_chess960);
	Move pondermove=m_ana.pondermove();
	if(pondermove.code()!=nullmove){
	  command+=" ponder " + pondermove.algebraic(m_chess960);
	}
	m_com.to_interface(command);
	m_ofs << "engine -> " << command << endl;
	m_ana.set_status(idle);
	//---------------------------------------------------------------------------------
      }else if(command=="quit"){
	set_action(stop_thinking);
	m_ana.quit();
      }
      m_command=m_com.from_interface();
    }
}


void Engine::to_console()
{
  string command;
  //------------------------------------ answer from engine -------------------------
  switch(m_ana.get_status())
    {
    case thinking:
      {
	if(m_mpv.version()<m_ana.mpv().version()){m_mpv=m_ana.mpv();send_info(m_ana);}
	break;
      }
    case finished:
      {
	if(m_mpv.version()<m_ana.mpv().version()){
	  m_mpv=m_ana.mpv();
	  m_mpv.set_version(m_ana.mpv().version());
	  send_info(m_ana);
	}
	if(m_status==e_thinking && m_status!=e_pondering && m_status!=e_infinite){
	  m_status=e_idle;
	  m_ana.set_status(idle);
	  send_hashfull(m_ana);
	  m_command="";
	  Move bestmove=m_ana.best_move();
	  m_previous_position+=" "+bestmove.algebraic(m_chess960);
	  cp.make_move(bestmove);
	  command="bestmove " + bestmove.algebraic(m_chess960);
	  Move pondermove=m_ana.pondermove();
	  if(pondermove.code()!=nullmove){
	    command+=" ponder " + pondermove.algebraic(m_chess960);
	  }
	  m_com.to_interface(command);
	  m_ofs << "engine -> " << command << endl;

	  command="";
	}
	break; 
      }
    }
  //------------------------------------ answer from engine -------------------------
}


#include"board/bposition.cpp"
#include"board/cposition.cpp"
#include"board/rep3filter.cpp"
#include"common/attacks.cpp"
#include"common/bishopmask.cpp"
#include"common/bitboard.cpp"
#include"common/connections.cpp"
#include"common/magicmoves.cpp"
#include"common/magicnumber.cpp"
#include"common/piece.cpp"
#include"common/rookmask.cpp"
#include"common/square.cpp"
#include"common/tables.cpp"
#include"communicator.cpp"
#include"console.cpp"
#include"egt/egt4pieces.cpp"
#include"eval/evaluation.cpp"
#include"eval/distance.cpp"
#include"eval/materialindex.cpp"
#include"eval/material.cpp"
#include"eval/mobility.cpp"
#include"eval/pst.cpp"
#include"board/hasher.cpp"
#include"hashtable/bucket.cpp"
#include"hashtable/pawnhashtable.cpp"
#include"hashtable/phtrecord.cpp"
#include"hashtable/transposition.cpp"
#include"hashtable/ttmove.cpp"
#include"hashtable/ttrecord.cpp"
#include"movegen/move.cpp"
#include"movegen/movegenerator.cpp"
#include"movegen/mlist.cpp"
#include"book/bkmove.cpp"
#include"book/bkrecord.cpp"
#include"book/book.cpp"
#include"search/analyze.cpp"
#include"search/countermove.cpp"
#include"search/duration.cpp"
#include"search/principal.cpp"
#include"search/history.cpp"
#include"search/killer.cpp"
#include"search/lmr.cpp"
#include"search/multipv.cpp"
#include"search/timemanager.cpp"
#include"search/timer.cpp"
