/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifndef EVALUATION_H
#define EVALUATION_H

#include"material.h"
#include"mobility.h"
#include"pst.h"
#include"distance.h"
#include"../hashtable/pawnhashtable.h"
#include"../egt/egt4pieces.h"
#include"weight.h"
#include<vector>
#include<string>
#include<thread>
#include<fstream>
using std::vector;
using std::string;
using std::thread;
using std::ofstream;


class ChessPosition;

class Evaluation
{
 public:
  Evaluation(const ChessPosition&,Tables&);
  ~Evaluation();

  void init();
  void clear();
  void make_move(const Move&);
  void unmake_move(const Move&);
  int static_eval();
  void set_log(bool);
  string fen()const;
  vector<float>& variable_vector();
  vector<float>& weight_mg_vector();
  vector<float>& weight_eg_vector();
  
  //----------- mobility: ---------------------------------------
  const BitBoard& white_king_neighbourhood()const{return mob.white_king_neighbourhood();}
  const BitBoard& black_king_neighbourhood()const{return mob.black_king_neighbourhood();}
  const BitBoard& white_king_battlefield()const{return mob.white_king_battlefield();}
  const BitBoard& black_king_battlefield()const{return mob.black_king_battlefield();}
 
  //-------- general (raw) values: ------------------------------
  int material()const;
  int mobility();
  int king_safety()const;
  int pair_of_bishops()const;
  BitBoard passed_pawns()const;

  void compute_phase();
  void compute_material();
  void compute_mobility();
  void compute_pawn_eval();
  void compute_development();
  void compute_space();
  void compute_pst();
  void compute_king_activity();
  void compute_king_safety();
  void compute_safe_queenchecks();
  void compute_pinned_pieces();
  void compute_opposite_colored_bishops();

  int distance_between(Square,Square)const;
  int distance_between(Square,BitBoard)const;

  string report();
  string stats();

 private:
  const ChessPosition& cp;
  PawnHashTable m_pht;
  Material m_mater;
  PieceSquareTables m_pst;
  Mobility mob;
  Distance m_d;
  PHTrecord m_phtr;
  Weight m_weight;
  vector<float> m_variable;

  //---------------------- stats: ---------------------------------------
  bool m_log{false};
  ofstream m_ofs_log;
  uint64_t m_eval_count{0};
  uint64_t m_new_pawn_eval{0};
  uint64_t m_missed_pht_read{0};
  //---------------------------------------------------------------------
  

  //------------------ bitboards: ---------------------------------------
  //
  BitBoard bb_white_passed_pawns,bb_black_passed_pawns;
  BitBoard bb_white_passed_pawn_stops,bb_black_passed_pawn_stops;
  BitBoard bb_white_passed_pawn_telestops,bb_black_passed_pawn_telestops;
  BitBoard bb_white_passed_pawn_rearspan,bb_black_passed_pawn_rearspan;
  BitBoard bb_white_weak_pawns,bb_black_weak_pawns;
  BitBoard bb_white_open_files,bb_black_open_files;
  //----------------------------------------------------------------------


  //-------------------- Numbers: -------------------------------------------
  //
  float m_middlegame{1};
  float m_endgame{0};
  int m_white_passed_pawns{0};
  int m_black_passed_pawns{0};
  int m_white_pawn_shield{0};
  int m_black_pawn_shield{0};

  // Final score:
  float m_score{0};
  //
  //-------------------- Numbers  -------------------------------------------

  
  //---------- Endgames: ------------------
  //
  int KQK();
  int KRK();
  int KRPKR()const;
  int KNNK()const;
  int KBBK()const;
  int KBNK();
  int KBPK()const;
  int KPK()const;
  int KQKR();
  //
};



#endif
