/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include<iostream>
#include<iomanip>
#include<stdexcept>
#include<cmath>
#include<cassert>
#include"pst.h"
#include"../board/cposition.h"
#include"../common/numerals.h"
using std::logic_error;
using std::cout;
using std::endl;
using std::round;
using std::hex;
using std::dec;




PieceSquareTables::PieceSquareTables()
{
  for(int i=0;i<64;i++){
    pst_mg[no_piece_id][i]=0;
    pst_mg[white_pawn_id][i]=pawn_mg[i];
    pst_mg[white_knight_id][i]=knight_mg[i];
    pst_mg[white_bishop_id][i]=bishop_mg[i];
    pst_mg[white_rook_id][i]=rook_mg[i];
    pst_mg[white_queen_id][i]=queen_mg[i];
    pst_mg[white_king_id][i]=king_mg[i];
    pst_mg[black_pawn_id][i]=-pawn_mg[i^56];
    pst_mg[black_knight_id][i]=-knight_mg[i^56];
    pst_mg[black_bishop_id][i]=-bishop_mg[i^56];
    pst_mg[black_rook_id][i]=-rook_mg[i^56];
    pst_mg[black_queen_id][i]=-queen_mg[i^56];
    pst_mg[black_king_id][i]=-king_mg[i^56];
  }

  for(int i=0;i<64;i++){
    pst_eg[no_piece_id][i]=0;
    pst_eg[white_pawn_id][i]=pawn_eg[i];
    pst_eg[white_knight_id][i]=knight_eg[i];
    pst_eg[white_bishop_id][i]=bishop_eg[i];
    pst_eg[white_rook_id][i]=rook_eg[i];
    pst_eg[white_queen_id][i]=queen_eg[i];
    pst_eg[white_king_id][i]=king_eg[i];
    pst_eg[black_pawn_id][i]=-pawn_eg[i^56];
    pst_eg[black_knight_id][i]=-knight_eg[i^56];
    pst_eg[black_bishop_id][i]=-bishop_eg[i^56];
    pst_eg[black_rook_id][i]=-rook_eg[i^56];
    pst_eg[black_queen_id][i]=-queen_eg[i^56];
    pst_eg[black_king_id][i]=-king_eg[i^56];
  }
}

void PieceSquareTables::init(const ChessPosition& cp)
{
  pcp=&cp;
  for(auto& s : m_score){s.first=0;s.second=0;}
  for(Square s=a1;s<=h8;s++){
    m_score[cp[s.id()]].first+=pst_mg[cp[s.id()]][s.id()];
    m_score[cp[s.id()]].second+=pst_eg[cp[s.id()]][s.id()];
  }
}

bool PieceSquareTables::verify_score()const
{
  array<pair<float,float>,15> score;
  for(auto& s : score){s.first=s.second=0;}
  for(Square s=a1;s<=h8;s++){
    score[pcp->operator[](s.id())].first+=pst_mg[pcp->operator[](s.id())][s.id()];
    score[pcp->operator[](s.id())].second+=pst_eg[pcp->operator[](s.id())][s.id()];
  }

  for(int i=0;i<15;i++){
    if(fabs(score[i].first - m_score[i].first) > 0.01){
      std::cerr << "verify correct score[" << i << "].first " << score[i].first << " != incremental score["<< i << "].first " << m_score[i].first << endl;
      std::cerr << "verify correct score[" << i << "].second " << score[i].second << " != incremental score["<< i << "].second " << m_score[i].second << endl;
      return false;
    }
    if(fabs(score[i].second - m_score[i].second) > 0.01){return false;}
  }

  return true;
}

void PieceSquareTables::make_move(const Move& m)
{
  int to{m.to().id()},from{m.from().id()};
  switch(m.code())
    {
    case nullmove:return;
    case normal_move:
      m_score[m.piece_id()].first+=(pst_mg[m.piece_id()][to]-pst_mg[m.piece_id()][from]);
      m_score[m.piece_id()].second+=(pst_eg[m.piece_id()][to]-pst_eg[m.piece_id()][from]);
      m_score[m.capture_id()].first-=pst_mg[m.capture_id()][to];
      m_score[m.capture_id()].second-=pst_eg[m.capture_id()][to];
      return;
    case en_passant:
      m_score[m.piece_id()].first+=(pst_mg[m.piece_id()][to]-pst_mg[m.piece_id()][from]);
      m_score[m.piece_id()].second+=(pst_eg[m.piece_id()][to]-pst_eg[m.piece_id()][from]);
      m_score[m.capture_id()].first-=pst_mg[m.capture_id()][to^8];
      m_score[m.capture_id()].second-=pst_eg[m.capture_id()][to^8];
      return;
    case castling:
      if(from<8){
	m_score[white_king_id].first-=pst_mg[white_king_id][from];
	m_score[white_king_id].second-=pst_eg[white_king_id][from];
	m_score[white_rook_id].first-=pst_mg[white_rook_id][to];
	m_score[white_rook_id].second-=pst_eg[white_rook_id][to];
	if(to>from){
	  m_score[white_king_id].first+=pst_mg[white_king_id][g1];
	  m_score[white_king_id].second+=pst_eg[white_king_id][g1];
	  m_score[white_rook_id].first+=pst_mg[white_rook_id][f1];
	  m_score[white_rook_id].second+=pst_eg[white_rook_id][f1];
	}else{
	  m_score[white_king_id].first+=pst_mg[white_king_id][c1];
	  m_score[white_king_id].second+=pst_eg[white_king_id][c1];
	  m_score[white_rook_id].first+=pst_mg[white_rook_id][d1];
	  m_score[white_rook_id].second+=pst_eg[white_rook_id][d1];
	}
      }else{
	m_score[black_king_id].first-=pst_mg[black_king_id][from];
	m_score[black_king_id].second-=pst_eg[black_king_id][from];
	m_score[black_rook_id].first-=pst_mg[black_rook_id][to];
	m_score[black_rook_id].second-=pst_eg[black_rook_id][to];
	if(to>from){
	  m_score[black_king_id].first+=pst_mg[black_king_id][g8];
	  m_score[black_king_id].second+=pst_eg[black_king_id][g8];
	  m_score[black_rook_id].first+=pst_mg[black_rook_id][f8];
	  m_score[black_rook_id].second+=pst_eg[black_rook_id][f8];
	}else{
	  m_score[black_king_id].first+=pst_mg[black_king_id][c8];
	  m_score[black_king_id].second+=pst_eg[black_king_id][c8];
	  m_score[black_rook_id].first+=pst_mg[black_rook_id][d8];
	  m_score[black_rook_id].second+=pst_eg[black_rook_id][d8];
	}
      }
      return;
    case promo_queen:
      promote<promo_queen>(m);
      return;
    case promo_rook:
      promote<promo_rook>(m);
      return;
    case promo_bishop:
      promote<promo_bishop>(m);
      return;
    case promo_knight:
      promote<promo_knight>(m);
      return;
    }
}

template<move_t T>
void PieceSquareTables::promote(const Move& m)
{
  int piece_id=T==promo_knight?white_knight_id:white_queen_id;
  piece_id=T==promo_bishop?white_bishop_id:piece_id;
  piece_id=T==promo_rook?white_rook_id:piece_id;
  int to{m.to().id()},from{m.from().id()};
  m_score[m.capture_id()].first-=pst_mg[m.capture_id()][to];
  m_score[m.capture_id()].second-=pst_eg[m.capture_id()][to];
  m_score[m.piece_id()].first-=pst_mg[m.piece_id()][from];
  m_score[m.piece_id()].second-=pst_eg[m.piece_id()][from];
  if(to>from){
    m_score[piece_id].first+=pst_mg[piece_id][to];
    m_score[piece_id].second+=pst_eg[piece_id][to];
  }else{
    piece_id^=8;
    m_score[piece_id].first+=pst_mg[piece_id][to];
    m_score[piece_id].second+=pst_eg[piece_id][to];
  }
  return;  
}


void PieceSquareTables::unmake_move(const Move& m)
{
  int to{m.to().id()},from{m.from().id()};
  switch(m.code())
    {
    case nullmove:return;
    case normal_move:
      m_score[m.piece_id()].first-=(pst_mg[m.piece_id()][to]-pst_mg[m.piece_id()][from]);
      m_score[m.piece_id()].second-=(pst_eg[m.piece_id()][to]-pst_eg[m.piece_id()][from]);
      m_score[m.capture_id()].first+=pst_mg[m.capture_id()][to];
      m_score[m.capture_id()].second+=pst_eg[m.capture_id()][to];
      return;
    case en_passant:
      m_score[m.piece_id()].first-=(pst_mg[m.piece_id()][to]-pst_mg[m.piece_id()][from]);
      m_score[m.piece_id()].second-=(pst_eg[m.piece_id()][to]-pst_eg[m.piece_id()][from]);
      m_score[m.capture_id()].first+=pst_mg[m.capture_id()][to^8];
      m_score[m.capture_id()].second+=pst_eg[m.capture_id()][to^8];
      return;
    case castling:
      if(from<8){
	m_score[white_king_id].first+=pst_mg[white_king_id][from];
	m_score[white_king_id].second+=pst_eg[white_king_id][from];
	m_score[white_rook_id].first+=pst_mg[white_rook_id][to];
	m_score[white_rook_id].second+=pst_eg[white_rook_id][to];
	if(to>from){
	  m_score[white_king_id].first-=pst_mg[white_king_id][g1];
	  m_score[white_king_id].second-=pst_eg[white_king_id][g1];
	  m_score[white_rook_id].first-=pst_mg[white_rook_id][f1];
	  m_score[white_rook_id].second-=pst_eg[white_rook_id][f1];
	}else{
	  m_score[white_king_id].first-=pst_mg[white_king_id][c1];
	  m_score[white_king_id].second-=pst_eg[white_king_id][c1];
	  m_score[white_rook_id].first-=pst_mg[white_rook_id][d1];
	  m_score[white_rook_id].second-=pst_eg[white_rook_id][d1];
	}
      }else{
	m_score[black_king_id].first+=pst_mg[black_king_id][from];
	m_score[black_king_id].second+=pst_eg[black_king_id][from];
	m_score[black_rook_id].first+=pst_mg[black_rook_id][to];
	m_score[black_rook_id].second+=pst_eg[black_rook_id][to];
	if(to>from){
	  m_score[black_king_id].first-=pst_mg[black_king_id][g8];
	  m_score[black_king_id].second-=pst_eg[black_king_id][g8];
	  m_score[black_rook_id].first-=pst_mg[black_rook_id][f8];
	  m_score[black_rook_id].second-=pst_eg[black_rook_id][f8];
	}else{
	  m_score[black_king_id].first-=pst_mg[black_king_id][c8];
	  m_score[black_king_id].second-=pst_eg[black_king_id][c8];
	  m_score[black_rook_id].first-=pst_mg[black_rook_id][d8];
	  m_score[black_rook_id].second-=pst_eg[black_rook_id][d8];
	}
      }
      return;
    case promo_queen:
      unpromote<promo_queen>(m);
      return;
    case promo_rook:
      unpromote<promo_rook>(m);
      return;
    case promo_bishop:
      unpromote<promo_bishop>(m);
      return;
    case promo_knight:
      unpromote<promo_knight>(m);
      return;
    }
}

template<move_t T>
void PieceSquareTables::unpromote(const Move& m)
{
  int piece_id=T==promo_knight?white_knight_id:white_queen_id;
  piece_id=T==promo_bishop?white_bishop_id:piece_id;
  piece_id=T==promo_rook?white_rook_id:piece_id;

  int to{m.to().id()},from{m.from().id()};
  m_score[m.capture_id()].first+=pst_mg[m.capture_id()][to];
  m_score[m.capture_id()].second+=pst_eg[m.capture_id()][to];
  m_score[m.piece_id()].first+=pst_mg[m.piece_id()][from];
  m_score[m.piece_id()].second+=pst_eg[m.piece_id()][from];
  if(to>from){
    m_score[piece_id].first-=pst_mg[piece_id][to];
    m_score[piece_id].second-=pst_eg[piece_id][to];
  }else{
    piece_id^=8;
    m_score[piece_id].first-=pst_mg[piece_id][to];
    m_score[piece_id].second-=pst_eg[piece_id][to];
  }
  return;  
}

pair<float,float> PieceSquareTables::get_score()const
{
  pair<float,float> r{0,0};
  for(auto& s : m_score){r.first+=s.first;r.second+=s.second;}
  return r;
}

pair<float,float> PieceSquareTables::get_pawn_score()const
{
  pair<float,float> r;
  r.first=m_score[white_pawn_id].first+m_score[black_pawn_id].first;
  r.second=m_score[white_pawn_id].second+m_score[black_pawn_id].second;
  return r;
}

pair<float,float> PieceSquareTables::get_knight_score()const
{
  pair<float,float> r;
  r.first=m_score[white_knight_id].first+m_score[black_knight_id].first;
  r.second=m_score[white_knight_id].second+m_score[black_knight_id].second;
  return r;
}

pair<float,float> PieceSquareTables::get_bishop_score()const
{
  pair<float,float> r;
  r.first=m_score[white_bishop_id].first+m_score[black_bishop_id].first;
  r.second=m_score[white_bishop_id].second+m_score[black_bishop_id].second;
  return r;
}

pair<float,float> PieceSquareTables::get_rook_score()const
{
  pair<float,float> r;
  r.first=m_score[white_rook_id].first+m_score[black_rook_id].first;
  r.second=m_score[white_rook_id].second+m_score[black_rook_id].second;
  return r;
}

pair<float,float> PieceSquareTables::get_queen_score()const
{
  pair<float,float> r;
  r.first=m_score[white_queen_id].first+m_score[black_queen_id].first;
  r.second=m_score[white_queen_id].second+m_score[black_queen_id].second;
  return r;
}

pair<float,float> PieceSquareTables::get_king_score()const
{
  pair<float,float> r;
  r.first=m_score[white_king_id].first+m_score[black_king_id].first;
  r.second=m_score[white_king_id].second+m_score[black_king_id].second;
  return r;
}
