/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include<iostream>
#include<iomanip>
#include<sstream>
#include"phtrecord.h"
#include"../common/prebb.h"
using std::ostringstream;
using std::endl;
using std::hex;
using std::dec;


const vector<char> PHTrecord::pawn_island_table=[]()->vector<char>
{
  vector<char> x(256);
  for(unsigned int i=0;i<256;++i)
    {
      unsigned int j=i;
      bool on_island=false;
      int n=0;
      for(int k=0;k<8;k++)
	{
	  if((j&0x01) && !on_island){++n;on_island=true;}
	  else if(!(j&0x01)){on_island=false;}	  
	  j>>=1;
	}
      x[i]=n;
    }
  return x;
}();


PHTrecord::PHTrecord()
{
  m_key=0;
}

void PHTrecord::clear()
{
  m_key=0;
  m_white_passed_pawns.clear();
  m_white_passed_pawn_stops.clear();
  m_white_passed_pawn_telestops.clear();
  m_white_passed_pawn_rearspan.clear();
  m_white_weak_pawns.clear();
  m_white_open_files.clear();
  m_black_passed_pawns.clear();
  m_black_passed_pawn_stops.clear();
  m_black_passed_pawn_telestops.clear();
  m_black_passed_pawn_rearspan.clear();
  m_black_weak_pawns.clear();
  m_black_open_files.clear();
  m_weighted_white_passed_pawns=0;
  m_weighted_black_passed_pawns=0;
  m_weighted_doubled_pawns=0;
  m_weak_pawns=0;
  m_isolated_pawns=0;
  m_pawn_islands=0;
  m_rook_pawns=0;
}

void PHTrecord::set_pawns(const BitBoard& wp,const BitBoard& bp)
{
  // compute open files:
  BitBoard bb;
  bb=wp;
  bb|=bb<<8;
  bb|=bb<<16;
  bb|=bb<<32;
  bb|=bb>>8;
  bb|=bb>>16;
  bb|=bb>>32;
  m_white_open_files=~bb;
  
  bb=bp;
  bb|=bb<<8;
  bb|=bb<<16;
  bb|=bb<<32;
  bb|=bb>>8;
  bb|=bb>>16;
  bb|=bb>>32;
  m_black_open_files=~bb;
  
  //--------------------- isolated pawns: ----------------------------
  //
  bb= bp & (m_black_open_files<<1 | a_file) & (m_black_open_files>>1 | h_file); 
  m_isolated_pawns=bb.popcount();
  bb= wp & (m_white_open_files<<1 | a_file) & (m_white_open_files>>1 | h_file); 
  m_isolated_pawns-=bb.popcount();

  //--------------------- weak pawns: --------------------------------
  //
  bb=((bp<<1 &= ~a_file) |= (bp>>1 &= ~h_file));
  bb|=bb>>8;
  bb|=bb>>16;
  bb|=bb>>32;
  m_black_weak_pawns = bp & ~bb;
  bb=((wp<<1 &= ~a_file) |= (wp>>1 &= ~h_file));
  bb|=bb<<8;
  bb|=bb<<16;
  bb|=bb<<32;
  m_white_weak_pawns = wp & ~bb;
  m_weak_pawns=m_black_weak_pawns.popcount() - m_white_weak_pawns.popcount();
  
  //-------------------- passed pawns: -------------------------------
  //
  bb=bp;
  bb|=(bb<<1 &= ~a_file);
  bb|=(bb>>1 &= ~h_file);
  bb=bb>>8;
  bb|=bb>>8;
  bb|=bb>>16;
  bb|=bb>>16;
  m_white_passed_pawns=wp&~bb;
  int p4=(m_white_passed_pawns & rank4).popcount();
  int p5=(m_white_passed_pawns & rank5).popcount();
  int p6=(m_white_passed_pawns & rank6).popcount();
  int p7=(m_white_passed_pawns & rank7).popcount();
  m_weighted_white_passed_pawns=p4*25+p5*50+p6*100+p7*200;

  bb=wp;
  bb|=(bb<<1 &= ~a_file);
  bb|=(bb>>1 &= ~h_file);
  bb=bb<<8;
  bb|=bb<<8;
  bb|=bb<<16;
  bb|=bb<<16;
  m_black_passed_pawns=bp&~bb;
  p4=(m_black_passed_pawns & rank5).popcount();  
  p5=(m_black_passed_pawns & rank4).popcount();  
  p6=(m_black_passed_pawns & rank3).popcount();  
  p7=(m_black_passed_pawns & rank2).popcount();
  m_weighted_black_passed_pawns=p4*25+p5*50+p6*100+p7*200;
  //
  //------------------------------------------------------------------

  
  //-------------------- connected passed pawns: ---------------------
  //
  BitBoard white_passed_pawn_files{m_white_passed_pawns};
  white_passed_pawn_files|=white_passed_pawn_files<<8;
  white_passed_pawn_files|=white_passed_pawn_files<<16;
  white_passed_pawn_files|=white_passed_pawn_files<<32;
  white_passed_pawn_files|=white_passed_pawn_files>>8;
  white_passed_pawn_files|=white_passed_pawn_files>>16;
  white_passed_pawn_files|=white_passed_pawn_files>>32;
  BitBoard white_connected_passed_pawn_files=
    (white_passed_pawn_files<<1 & white_passed_pawn_files & ~a_file)
    | (white_passed_pawn_files>>1 & white_passed_pawn_files & ~h_file);
  BitBoard white_connected_passed_pawns=white_connected_passed_pawn_files & m_white_passed_pawns;

  BitBoard black_passed_pawn_files{m_black_passed_pawns};
  black_passed_pawn_files|=black_passed_pawn_files<<8;
  black_passed_pawn_files|=black_passed_pawn_files<<16;
  black_passed_pawn_files|=black_passed_pawn_files<<32;
  black_passed_pawn_files|=black_passed_pawn_files>>8;
  black_passed_pawn_files|=black_passed_pawn_files>>16;
  black_passed_pawn_files|=black_passed_pawn_files>>32;
  BitBoard black_connected_passed_pawn_files=
    (black_passed_pawn_files<<1 & black_passed_pawn_files & ~a_file)
    | (black_passed_pawn_files>>1 & black_passed_pawn_files & ~h_file);
  BitBoard black_connected_passed_pawns=black_connected_passed_pawn_files & m_black_passed_pawns;

  p4=(white_connected_passed_pawns & rank4).popcount();
  p5=(white_connected_passed_pawns & rank5).popcount();
  p6=(white_connected_passed_pawns & rank6).popcount();
  p7=(white_connected_passed_pawns & rank7).popcount();
  m_weighted_white_passed_pawns+=p4*25+p5*50+p6*100+p7*200;
  p4=(black_connected_passed_pawns & rank5).popcount();  
  p5=(black_connected_passed_pawns & rank4).popcount();  
  p6=(black_connected_passed_pawns & rank3).popcount();  
  p7=(black_connected_passed_pawns & rank2).popcount();  
  m_weighted_black_passed_pawns+=p4*25+p5*50+p6*100+p7*200;

  
  //------------------------ telestops: -------------------------
  //
  m_white_passed_pawn_stops=m_white_passed_pawns<<8;
  m_white_passed_pawn_telestops=m_white_passed_pawn_stops<<8;
  m_white_passed_pawn_telestops|=m_white_passed_pawn_telestops<<8;
  m_white_passed_pawn_telestops|=m_white_passed_pawn_telestops<<16;
  m_white_passed_pawn_telestops|=m_white_passed_pawn_telestops<<16;
  m_black_passed_pawn_stops=m_black_passed_pawns>>8;
  m_black_passed_pawn_telestops=m_black_passed_pawn_stops>>8;
  m_black_passed_pawn_telestops|=m_black_passed_pawn_telestops>>8;
  m_black_passed_pawn_telestops|=m_black_passed_pawn_telestops>>16;
  m_black_passed_pawn_telestops|=m_black_passed_pawn_telestops>>16;
  //
  //----------------------------------------------------------------
  
  //------------------- passed pawn rearspan --------------------
  //
  m_white_passed_pawn_rearspan=white_passed_pawn_files & ~m_white_passed_pawn_stops &  ~m_white_passed_pawn_telestops & ~m_white_passed_pawns; 
  m_black_passed_pawn_rearspan=black_passed_pawn_files & ~m_black_passed_pawn_stops &  ~m_black_passed_pawn_telestops & ~m_black_passed_pawns; 
  //
  //----------------------------------------------------------------
  
  //-------------------------- pawn islands: -----------------------
  //
  unsigned char c1=BitBoard::to_ull(~m_white_open_files&rank1);
  unsigned char c2=BitBoard::to_ull(~m_black_open_files&rank1);
  m_pawn_islands=pawn_island_table[c2]-pawn_island_table[c1];
  //
  //----------------------------------------------------------------
  
  //-------------------------- doubled pawns: ----------------------
  //
  bb=((wp>>8 &= wp) |= (wp>>16 &= wp) |= (wp>>24 &= wp) |= (wp>>32 &= wp));
  m_weighted_doubled_pawns=-bb.popcount();  
  bb=((bp<<8 &= bp) |= (bp<<16 &= bp) |= (bp<<24 &= bp) |= (bp<<32 &= bp));
  m_weighted_doubled_pawns+=bb.popcount();
  //
  //----------------------------------------------------------------
  
  //-------------------------- rook pawns: -------------------------
  //
  m_rook_pawns=-(wp & (a_file | h_file)).popcount();
  m_rook_pawns+=(bp & (a_file | h_file)).popcount();
  //
  //----------------------------------------------------------------


}
