/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../board/cposition.h"
#include"../eval/evaluation.h"
#include"killer.h"
#include"analyze.h"
#include"../common/numerals.h"


#include<algorithm>
#include<thread>
#include<future>
#include<ratio>
#include<chrono>
#include<stdexcept>
#include<iostream>
#include<iomanip>
#include<utility>
#include<sstream>

using std::to_string;
using std::back_inserter;
using std::hex;
using std::dec;
using std::cout;
using std::flush;
using std::endl;
using std::pair;
using std::boolalpha;
using std::logic_error;
using std::istringstream;
using std::ostringstream;
using std::setprecision;
using std::setw;
using std::async;
using std::launch;
using namespace std::chrono;



Analysis::Analysis(ChessPosition& x,Tables& y): cp(x), m_tables(y), m_tt(y.transpositiontable()), m_mg(MoveGenerator(y)),
						m_egtkbnk(Egt4Pieces(y.attacks(),white_knight_id,white_bishop_id)),
						m_egtkqkr(Egt4Pieces(y.attacks(),white_queen_id,black_rook_id)),
						m_lmr(y.lmr()), m_eva(x,y)
{
  m_ply=0;
  m_action=no_action;
  m_status=idle;
  m_quit=false;
  init(cp);
  m_tt.clear();
}

Analysis::Analysis(const Analysis& x, ChessPosition& y): cp(y), m_tables(x.m_tables), m_tt(x.m_tt), m_mg(MoveGenerator(x.m_tables)), m_egtkbnk(x.m_egtkbnk), m_egtkqkr(x.m_egtkqkr),
							 m_lmr(x.m_lmr), m_eva(y,x.m_tables), m_ply(x.m_ply), m_date(x.m_date)
{
  m_status=idle;
  m_quit=false;
  int tmp=x.m_iteration;
  m_iteration=tmp;
  tmp=x.m_seldepth;
  m_seldepth=tmp;
  init(cp);
}

Analysis::~Analysis()
{
}

void Analysis::init(const ChessPosition& x)
{
  if(m_status==thinking){
    throw logic_error("Analysis::init  init() call while thinking not allowed");
  }
  cp=x;
  m_action=no_action;
  m_status=idle;
  m_eva.init();
  m_tman.reset();
}

void Analysis::reset()
{
  m_sd_depth=m_max_depth;
  m_searchmoves.clear();
  m_tman.reset();
}

void Analysis::set_wtime(int x)
{
  m_tman.set_wtime(x);
}

void Analysis::set_btime(int x)
{
  m_tman.set_btime(x);
}

void Analysis::set_winc(int x)
{
  m_tman.set_winc(x);
}

void Analysis::set_binc(int x)
{
  m_tman.set_binc(x);
}

void Analysis::set_movestogo(int x)
{
  m_tman.set_white_to_move(cp.stm());
  m_tman.set_movestogo(x);
}

void Analysis::set_depth(int x)
{
  m_sd_depth=x;
}

void Analysis::set_nodes(uint64_t x)
{
  m_max_nodes=x;
}

void Analysis::set_mate(int x)
{
  m_sd_depth=x;
}

void Analysis::set_movetime(int x)
{
  m_tman.set_movetime(x);
}

void Analysis::set_infinite(bool x)
{
  m_tman.set_infinite(x);
}

void Analysis::set_pondering(bool x)
{
  m_tman.set_pondering(x);
}

void Analysis::set_searchmoves(vector<Move> x)
{
  m_tt.clear();
  m_searchmoves=x;
}

void Analysis::set_uci_limitstrength(bool x)
{
  m_uci_limitstrength=x;
  if(!x){
    m_elo_maxdepth=max_search_depth;
    m_handicap=0.0;
  }
}

void Analysis::set_uci_elo(int x)
{
  if(x<1500){x=1500;}
  if(x>2200){x=2200;}
  m_uci_elo=x;
  if(x<1600){m_elo_maxdepth=2;}
  else if(x<1700){m_elo_maxdepth=4;}
  else if(x<1800){m_elo_maxdepth=6;}
  else if(x<1900){m_elo_maxdepth=8;}
  else if(x<2000){m_elo_maxdepth=10;}
  else if(x<2100){m_elo_maxdepth=12;}
  else if(x<2200){m_elo_maxdepth=14;}
  if(x>=2000){m_handicap=0;}
  if(x<2000){m_handicap=200*(2000-x)/500;}
}

void Analysis::clear()
{
  clear_eval();
  clear_hash();
  clear_killer();
  clear_history();
  clear_node_count();
}

void Analysis::clear_eval()
{
  m_eva.clear();
}

void Analysis::clear_hash()
{
  m_tt.clear();
}

void Analysis::hashtable_enabled(bool x)
{
  m_tt.set_enabled(x);
}

void Analysis::clear_killer()
{
  m_killer.clear();
}

void Analysis::clear_history()
{
  m_ht.clear();
  m_cm.clear();
}

void Analysis::set_lag(int x)
{
  m_tman.set_lag(x);
}

void Analysis::set_log(bool x)
{
  m_log=x;
  ostringstream oss;
  oss << "analysis_" << cp.id() << ".log";
  if(x){
    ofs_log.open(oss.str());
  }else{
    ofs_log.close();
  }
  m_eva.set_log(x);
}

void Analysis::quit()
{
  m_quit=true;
}

void Analysis::set_action(action_t x)
{
  m_action=x;
}

int Analysis::get_status()const
{
  return m_status;
}

void Analysis::set_status(int x)
{
  m_status=x;
}

void Analysis::set_id(int x)
{
  m_id=x;
}

int Analysis::get_id()const
{
  return m_id;
}

void Analysis::ponderhit()
{
  m_tman.set_pondering(false);
}

void Analysis::new_node()
{
  if(m_max_nodes!=infinite_nodes){
    if(m_node_count>=m_max_nodes){
      m_action=stop_thinking;
      return;
    }
  }else if((m_node_count&0x3FFF)==0){
    if(m_ply>2 && m_tman.time_to_move()){
      m_action=stop_thinking;
      return;
    }
  }
  m_node_count++;
  return;
}

int Analysis::root_search(MoveList& ml,int depth,int alpha,int beta)
{
  //------------------------- EGT ? ----------------------------------------------------------------------
  {
    bool egt{false};
    if(!cp.white_pawn() && !cp.black_pawn()){
      if((cp.white_queen() | cp.black_queen()
	  | cp.white_rook() | cp.black_rook()
	  | cp.white_bishop() | cp.black_bishop()
	  | cp.white_knight() | cp.black_knight()).popcount()<=2){      
	if(cp.materialindex().str()=="KBNk" || cp.materialindex().str()=="Kkbn"){
	  //
	  egt=true;
	  if(!m_egtkbnk.ready()){
	    m_tt.set_threshold(6000000);
	    if(!m_egtkbnk.busy()){m_egtkbnk.set_memory(m_tt.get_uchar_adress());}
	  }
	  while(m_tman.remaining_time()>2000){
	    m_egtkbnk.create();
	    if(m_egtkbnk.ready()){
	      break;
	    }
	  }
	  //
	}else if(cp.materialindex().str()=="KQkr" || cp.materialindex().str()=="KRkq" || cp.materialindex().str()=="KQk"
		 || cp.materialindex().str()=="Kkq" || cp.materialindex().str()=="KRk" || cp.materialindex().str()=="Kkr"){
	  //
	  egt=true;
	  if(!m_egtkqkr.ready()){
	    m_tt.set_threshold(6000000);
	    if(!m_egtkqkr.busy()){
	      m_egtkqkr.set_memory(m_tt.get_uchar_adress());
	    }
	  }
	  while(m_tman.remaining_time()>4000){
	    m_egtkqkr.create();
	    if(m_egtkqkr.ready()){
	      break;
	    }
	  }
	  //
	}
      }
    }
    if(!egt){
      m_egtkqkr.clear();
      m_egtkbnk.clear();
      m_tt.set_threshold(0);
    }
  }
  //------------------------------------------------------------------------------------------------------

  PrincipalVariation pv;
  int eval=pv_search(depth,alpha,beta,pv);
  m_mutex_mpv.lock();
  m_mpv.clear();
  m_mpv.add(m_pv);
  m_mutex_mpv.unlock();

  if(m_mpv.number()>1){
    auto old_value=m_value;
    auto old_best_move=m_best_move;

    PrincipalVariation best_pv;
    Move best_move;
    Move aml[256];
    MoveList ml(aml);
    m_mg.generate_moves(cp,ml);
    remove(m_best_move,ml);
    for(int i=1;i<m_mpv.number();i++){  
      alpha=-infinite_value;
      beta=infinite_value;
      for(size_t j=0;j<ml.size();j++){
	make_move(ml[j]);
	int v=-pv_search(depth-1,alpha,beta,pv);
	unmake_move(ml[j]);
	if(v>alpha){
	  alpha=v;
	  best_move=ml[j];
	  best_pv=ml[j]+pv;
	  best_pv.set_score(alpha);
	  best_pv.set_depth(depth);
	}
      }
      m_mutex_mpv.lock();
      m_mpv.add(best_pv);
      m_mutex_mpv.unlock();
      remove(best_move,ml);      
    }
    
    m_value=old_value;
    m_best_move=old_best_move;
  }
  
  return eval;
}

int Analysis::pv_search(const int& depth,int alpha,int beta,PrincipalVariation& pv)
{
  new_node();


  //------------------------------------ mate distance pruning --------------------------------
  //
  int mating_value=mate_value-m_ply;
  if(mating_value < beta){
    beta=mating_value;
    if(alpha>=beta){
      return beta;
    }
  }
  if(alpha < -mating_value){
    alpha=-mating_value;
    if(alpha>=beta){
      return alpha;
    }
  }
  //
  //-------------------------------------------------------------------------------------------
  
  array<uint64_t,64> bb_moves;
  m_mg.create_movetable(cp,bb_moves);
  
  if(m_ply>0){
    if(cp.draw_by_threefold_repetition()){return 0;}
    if(cp.get_dfz()>100){return 0;}
    //---------------------- egt pruning ----------------------------
    //
    if(!cp.check()){
      int static_value=static_eval();
      if(abs(static_value)>mate_value-max_search_depth){
	return static_value;
      }
    }
    //
    //---------------------------------------------------------------
  }
  pv.clear();
  if(m_ply>m_seldepth){m_seldepth=m_ply;}
  if(m_action==stop_thinking){return alpha;}
  bound_t cutoff=upperbound;
  PrincipalVariation child_pv;
  Move best_move;
  int value{-infinite_value};
  
  if(depth<=0 || m_ply>=max_search_depth){
    m_leave_nodes++;
    value=quiescence(alpha,beta,0);
    if(cp.get_dfz()>50){
      double f=1.0-0.02*(cp.get_dfz()-50.0);
      f=min(1.0,f);
      value*=f;
    }
    return value; 
  }


  //----------------------- hashtable move ------------------------
  //
  TTrecord tt;
  tt.set_key(cp.get_hash());
  if(m_tt.read(tt)){
    TTmove ttm=tt.get_move();
    if(is_legal(ttm)){
      best_move=ttm(cp);
    }
  }
  //
  //----------------------- hashtable move ------------------------


  Move aml[256];
  MoveList ml(aml);
  m_mg.serialize_moves(cp,bb_moves,ml);
  select_searchmoves(ml);
    
  if(ml.size()==0){
    if(!cp.check()){return 0;}else{return -mating_value;}
  }else{
    if(m_iteration==1 && m_ply==0){m_best_move=ml[0];}
  }
  localsort(ml,depth,true,best_move);

    
  Move m;
  for(size_t mi=0;mi<ml.size();mi++){
    m=ml[mi];

      
    //----------------- extensions: ---------------------------------------
    //
    int extension=extend(ml,m);
    //
    //---------------------------------------------------------------------

    //------------------- LMR: --------------------------------------------
    //
    int lmr{0};	
    //if( extension==0 && !cp.check() && m.xcode()!=good_capture && !m.check() && m.xcode()!=killer && !m.piece().king() ){
    if( extension==0 && !cp.check() && m.xcode()!=good_capture && !m.check() && m.xcode()!=killer){
      lmr=m_lmr.pv_value(depth,mi);
    }
    //
    //---------------------------------------------------------------------

    make_move(m);
    if(mi==0){               // principal variation search
      value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
    }else{
      value=-nullwindow_search(depth-1-lmr+extension,-alpha-1,-alpha);
      if(value>alpha){
	if(lmr>0){           // repeat null window search without lmr
	  value=-nullwindow_search(depth-1+extension,-alpha-1,-alpha);
	  if(value>alpha){   // repeat search with open window
	    value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
	  }
	}else{
	  value=-pv_search(depth-1+extension,-beta,-alpha,child_pv);
	}
      }
    }
    unmake_move(m);
    if(m_action==stop_thinking){return alpha;}
	
    if(value>=beta){
      put_in_tt(lowerbound,depth,beta,m,m_date);
      return beta;
    }
    if(value>alpha){
      best_move=m;
      alpha=value;
      cutoff=exact_value;
      pv.clear();
      pv=best_move+child_pv;
      if(m_ply==0){
	pv.set_depth(depth);
	pv.set_score(alpha);
	m_pv=pv;
	m_value=alpha;
	m_best_move=best_move;
	if(m_tman.percentage()>60){
	  set_action(stop_thinking);
	}
      }
    }else if(mi==0){
      iidsort(ml,depth,alpha,beta);
    }
  }
    
  put_in_tt(cutoff,depth,alpha,best_move,m_date);



  return alpha;
}

int Analysis::nullwindow_search(int depth,int alpha,int beta)
{
  new_node();

  
  //------------------------------------ mate distance pruning --------------------------------
  //
  int mating_value=mate_value-m_ply;
  if(mating_value < beta){
    beta=mating_value;
    if(alpha>=beta){
      return beta;
    }
  }
  if(alpha < -mating_value){
    alpha=-mating_value;
    if(alpha>=beta){
      return alpha;
    }
  }
  //
  //-------------------------------------------------------------------------------------------
  
  if(m_ply>m_seldepth){m_seldepth=m_ply;}
  if(m_action==stop_thinking){return alpha;}
  if(cp.draw_by_threefold_repetition()){return 0;}
  bound_t cutoff=upperbound;
  Move best_move,hashtable_move;
  int value{-infinite_value};
  
  if(depth<=0 || m_ply>=max_search_depth)
    {
      m_leave_nodes++;
      value=quiescence(alpha,beta,0);
      if(cp.get_dfz()>50){
	double f=1.0-0.02*(cp.get_dfz()-50.0);
	f=min(1.0,f);
	value*=f;
      }
      return value;
    }
  
  
  array<uint64_t,64> bb_moves;
  m_mg.create_movetable(cp,bb_moves);

  
  int static_value;
  if(!cp.check()){
    static_value=static_eval();
    //---------------------- egt pruning ----------------------------
    //
    if(static_value>mate_value-max_search_depth || static_value<-mate_value+max_search_depth){
      return static_value;
    }
    //
    //---------------------------------------------------------------
  }else{
    static_value=-infinite_value;
  }
  
  //---------------------- nullmove pruning -----------------------
  //
  if(static_value>=beta && nullmove_allowed()){
    value=nullmove_search(depth,alpha,beta);
    if(value>=beta){
      return beta;
    }else{
      value=nullmove_search(depth,alpha-2*pawn_value,beta-2*pawn_value);
      if(value<alpha-2*pawn_value){
	depth+=1;
      }
    }
  }
  //
  //---------------------- nullmove pruning -----------------------

      
  //------------------------- probcut -----------------------------
  //
  int pb=beta+2*pawn_value;
  if(static_value>=pb && !cp.check() && depth>4 && m_mate_search_level==0){
    value=nullwindow_search(depth-4,pb-1,pb);
    if(value>=pb){
      return beta;
    }
  }
  //
  //---------------------------------------------------------------

      

  //----------------------- hashtable move ------------------------
  //
  TTrecord tt;
  tt.set_key(cp.get_hash());
  if(m_tt.read(tt)){
    TTmove ttm=tt.get_move();
    auto score=tt.get_score(m_ply);
    if(tt.get_draft()>=depth){
      if(tt.is_exact_value()){
	return score;
      }else if(tt.is_upperbound() && score<=alpha){
	return alpha;
      }else if(tt.is_lowerbound() && score>=beta){
	return beta;
      }
    }
    if(is_legal(ttm)){
      best_move=ttm(cp);

      MoveList ml;
      int extension=extend(ml,best_move);

      make_move(best_move);
      value=-nullwindow_search(depth-1+extension,-beta,-alpha);
      unmake_move(best_move);
      if(value>=beta){
	put_in_tt(lowerbound,depth,beta,best_move,m_date);
	return beta;
      }
    }
  }
  //
  //----------------------- hashtable move ------------------------

  uint32_t aml[256];
  MoveList ml(reinterpret_cast<Move*>(aml)); // (avoid Move constructor)
  if(best_move.code()!=nullmove){m_mg.remove(best_move,bb_moves);}
  m_mg.serialize_moves(cp,bb_moves,ml);
      
  if(ml.size()==0 && best_move.code()==nullmove){if(!cp.check()){return 0;}else{return -mating_value;}}
  
  localsort(ml,depth,false);
      
        
  Move m;
  for(size_t mi=0;mi<ml.size();mi++){
    m=ml[mi];
    if(m==best_move){continue;}
	
    if(!cp.check()){
      //---------------------- LMP (late move pruning) ------------------
      //
      if(m_ply>4){
	if((depth<=2 && mi>6) || (depth<=4 && mi>10)){
	  if(!m.capture() && m.code()==normal_move && !m.check()){
	    continue;
	  }
	}
      }
      //
      //---------------------------------------------------------------
	
      //------------------ futility pruning: --------------------------------
      //
      if(!m.capture() && m.code()==normal_move && !m.check()){
	if((depth==1 && static_value <= alpha - pawn_value) ||
	   (depth==2 && static_value <= alpha - rook_value) ||
	   (depth==3 && static_value <= alpha - queen_value)){
	  continue;
	}
      }
      //
      //---------------------------------------------------------------------
    }	
	
    //----------------- extensions: ---------------------------------------
    //
    int extension=extend(ml,m);
    //
    //---------------------------------------------------------------------
	
    //------------------- LMR: --------------------------------------------
    //
    int lmr{0};
    if(extension==0 &&  !cp.check() &&  m.xcode()!=good_capture && !m.check() && m.xcode()!=killer){
      lmr=m_lmr.value(depth,mi);
    }
    //
    //---------------------------------------------------------------------
    

    //----------------------------------------------------------------------------------------------------------------------------------------
    //
    make_move(m);
    value=-nullwindow_search(depth-1-lmr+extension,-beta,-alpha);
    if(value>=beta && lmr>0){
      value=-nullwindow_search(depth-1,-beta,-alpha);              // re-search without LMR
    }
    unmake_move(m);
    if(m_action==stop_thinking){return alpha;}
    //
    //----------------------------------------------------------------------------------------------------------------------------------------

	
    if(value>alpha){
      best_move=m;
      alpha=beta;
      cutoff=lowerbound;
      if(!best_move.capture()){
	// countermove:
	size_t hs{cp.history().size()};
	if(hs>0 && cp.history().back().last_move!=NullMove()){
	  m_cm.increment(cp.history().back().last_move,best_move,depth);
	  for(unsigned i=0;ml[i]!=m;i++){
	    m_cm.decrement(cp.history().back().last_move,ml[i],depth);
	  }
	  if(hs>1 && cp.history()[hs-1].last_move!=NullMove()  && cp.history()[hs-2].last_move!=NullMove()){
	    m_cm.increment(cp.history()[hs-2].last_move,best_move,depth);
	    for(unsigned i=0;ml[i]!=m;i++){
	      m_cm.decrement(cp.history()[hs-2].last_move,ml[i],depth);
	    }
	  }
	}
	// history table:
	m_ht.increment(best_move,depth);
	for(unsigned i=0;ml[i]!=m;i++){
	  m_ht.decrement(ml[i],depth);
	}
	// killers:
	m_killer.set_killer(m_ply,best_move);
      }
      break;
    }
  }

  put_in_tt(cutoff,depth,alpha,best_move,m_date);


  return alpha;
}

int Analysis::nullmove_search(int depth,int alpha,int beta)
{
  int R=depth/5+3;
  make_nullmove();
  m_nullmove_level++;
  int value=-nullwindow_search(max(0,depth-R-1),-beta,-alpha);
  m_nullmove_level--;
  unmake_nullmove();
  return value;
}

int Analysis::extend(const MoveList& ml,const Move& m)
{
  BitBoard passed_pawns=m_eva.passed_pawns();
  BitBoard white_king_battlefield=m_eva.white_king_battlefield();
  BitBoard black_king_battlefield=m_eva.black_king_battlefield();
  int extension=0;
  if(m_ply <= m_iteration+8){
    if(m.check() && see(m)>=0){extension++;}                                      // check extension
    else if(ml.size()==1){extension++;}                                           // single move extension
    else if(m.piece().white_pawn() && m.to().rank_number()==7){extension++;}      // passed pawn extension
    else if(m.piece().black_pawn() && m.to().rank_number()==2){extension++;}      // passed pawn extension
    else if(m.xcode()==mate_attack){extension++;}                                 // mate search extension

    if(m.capture() && cp.history().size()>0 && cp.history().back().last_move.capture() && m.to()==cp.history().back().last_move.to()){
      if((cp.stm()==white && (m.to().bitboard() & black_king_battlefield)) ||
	 (cp.stm()==black && (m.to().bitboard() & white_king_battlefield))){
	extension++;                                                              // recapture extension in king battlefield 
      }
    }
    
    if(extension==0 && cp.nobelmen()<=2*rook_value){
      if(m.piece().pawn() && (m.from().bitboard() & passed_pawns)){extension++;}  // passed pawn race extension
    }
    
  }
  return extension;
}

void Analysis::iterative_search()
{

  m_eva.init();
  m_tman.start();
  m_tm.start();
  m_seldepth=0;
  m_node_count=0;
  m_leave_nodes=0;
  m_mate_search_level=0;
  m_best_move.clear();
  m_pondermove.clear();
  m_value=0;
  m_ht.divide_by_8();
  m_cm.divide_by_8();
  m_killer.clear();
  m_pv.clear();
  m_date++;
  int aspiration_alpha=-infinite_value;
  int aspiration_beta=infinite_value;
  m_iteration=0;
  MoveList ml;
  while(m_iteration<m_max_depth && m_iteration<m_sd_depth && abs(m_value) < (mate_value - m_iteration) && (!m_uci_limitstrength || m_iteration<=m_elo_maxdepth)){
    m_iteration++;
    if(m_uci_limitstrength){m_tt.clear();}
    m_ply=0;
    int alpha=aspiration_alpha,beta=aspiration_beta;
    int v=root_search(ml,m_iteration,alpha,beta);
    if(m_mpv.size()>0){
      if(m_mpv[0].size()>0){m_best_move=m_mpv[0][0];}
      if(m_mpv[0].size()>1){m_pondermove=m_mpv[0][1];}else{m_pondermove.clear();}
    }
    
    //------------------------ aspiration window handling: --------------------------------------
    //
    if(m_baspiration && m_action!=stop_thinking && m_mpv.number()==1){
      if(v<=aspiration_alpha && aspiration_alpha!=-infinite_value){
	aspiration_alpha-=aspiration_beta-aspiration_alpha;
	if(aspiration_alpha<-rook_value){aspiration_alpha=-infinite_value;}
	m_iteration--;
	continue;
      }else if(v>=aspiration_beta && aspiration_beta!=infinite_value){
	aspiration_beta+=aspiration_beta-aspiration_alpha;
	if(aspiration_beta>rook_value){aspiration_beta=infinite_value;}
	m_iteration--;
	continue;
      }else{
	aspiration_alpha=v-pawn_value/4;
	aspiration_beta=v+pawn_value/4;
	if(aspiration_alpha<-rook_value){aspiration_alpha=-infinite_value;}
	if(aspiration_beta>rook_value){aspiration_beta=infinite_value;}
      }
    }
    //
    //------------------------ aspiration window handling  --------------------------------------
    
    
    if(m_action!=stop_thinking){
      float exponent=1.0/m_iteration;
      m_ebf=pow(m_leave_nodes,exponent);
    }
    m_leave_nodes=0;

    if(m_action==stop_thinking){break;}
    //-------------------- end iteration ---------------------------------
  }

  if(m_uci_limitstrength){
    m_mpv.randomize(m_handicap);
    m_best_move=m_mpv[0][0];
    m_pondermove=NullMove();
    while(m_action!=stop_thinking && !m_tman.time_to_move()){
      std::this_thread::sleep_for(milliseconds(50));      
    }
  }
  
  m_tm.stop();
}

bool Analysis::nullmove_allowed()const
{
  if(cp.last_move().code()==nullmove){return false;}
  if(cp.check()){return false;}
  if(m_ply<1){return false;}
  if(cp.stm()==white){
    if(cp.white_pieces() & ~cp.white_king() & ~cp.white_knight() & ~cp.white_pawn()){return true;}
    else{return false;}
  }else{
   if(cp.black_pieces() & ~cp.black_king() & ~cp.black_knight() & ~cp.black_pawn()){return true;}
    else{return false;}
  }
}

bool Analysis::is_legal(TTmove& m)const
{
  if(m.from()==m.to())return false;
  if(cp.stm()==white){
    if(!(m.from().bitboard()&cp.white_pieces())){return false;}
    if(m.to().bitboard()&cp.black_king()){return false;}
  }else{
    if(!(m.from().bitboard()&cp.black_pieces())){return false;}
    if(m.to().bitboard()&cp.white_king()){return false;}
  }
  return true;
}

void Analysis::set_multipv(unsigned int n)
{
  if(n<1){n=1;}
  if(n>50){n=50;}
  m_mpv.set_number(n);
}

void Analysis::set_max_threads(unsigned int n)
{
  if(n<1){n=1;}
  if(n>2){n=2;}
  m_max_threads=n;
}

void Analysis::set_max_nodes(uint64_t n)
{
  m_max_nodes=n;
}

void Analysis::reset_max_nodes()
{
  m_max_nodes=infinite_nodes;
}

int Analysis::eval()const
{
  return m_value;
}

uint64_t Analysis::get_node_count()const
{
  return m_node_count;
}

void Analysis::add_node_count(uint64_t n)
{
  m_node_count+=n;
}

void Analysis::clear_node_count()
{
  m_node_count=0;
}

uint64_t Analysis::get_running_time()const
{
  return m_tm.running_time().milliseconds();
}

float Analysis::get_ebf()const
{
  return m_ebf;
}

int Analysis::get_hashfull()const
{
  return m_tt.hashfull(m_date);
}

int Analysis::get_iteration()const
{
  int r=m_iteration;
  return r;
}

int Analysis::get_seldepth()const
{
  int r=m_seldepth;
  return r;
}

void Analysis::set_hash_size(int i)
{
  m_tt.set_size(i);
}

void Analysis::set_aspiration(bool x)
{
  m_baspiration=x;
}

void Analysis::select_searchmoves(MoveList& ml)
{
  if(m_ply==0 && m_searchmoves.size()>0){
    int i{0};
    Move vttm[256];
    for(size_t j=0;j<ml.size();j++){
      for(const auto& sm : m_searchmoves){
	if(ml[j]==sm){vttm[i++]=ml[j];}
      }
    }
    ml.set_size(i);
    for(size_t j=0;j<ml.size();j++){ml[j]=vttm[j];}
  }
}

void Analysis::remove(const Move& m,MoveList& ml)
{
  size_t size=ml.size();
  for(size_t i=0;i<size;i++){
    if(m==ml[i]){
      for(size_t j=i;j<size-1;j++){
	ml[j]=ml[j+1];
      }
      ml.set_size(size-1);
      break;
    }
  }
}

const ChessPosition& Analysis::chessposition()const
{
  return cp;
}

PrincipalVariation Analysis::pv()const
{
  m_mutex_mpv.lock();
  PrincipalVariation r{m_mpv[0]};
  m_mutex_mpv.unlock();
  return r;
}

MultiPV Analysis::mpv()const
{
  m_mutex_mpv.lock();
  MultiPV r{m_mpv};
  m_mutex_mpv.unlock();
  return r;
}

Move Analysis::best_move()const
{
  return m_best_move;
}

Move Analysis::pondermove()const
{
  return m_pondermove;
}

void Analysis::make_move(const Move& m)
{
  cp.make_move(m);
  m_eva.make_move(m);
  ++m_ply;
  if(m.xcode()==mate_attack){
    m_mate_search_level++;
  }
}

void Analysis::unmake_move(const Move& m)
{
  cp.unmake_move();
  m_eva.unmake_move(m);
  --m_ply;
  if(m.xcode()==mate_attack){
    m_mate_search_level--;
  }
}

void Analysis::make_nullmove()
{
  NullMove m;
  cp.make_move(m);
  ++m_ply;
}

void Analysis::unmake_nullmove()
{
  cp.unmake_move();--m_ply;
}

void Analysis::operator()()
{
  while(true)
    {
      if(m_action==start_iterations){
	m_action=no_action;
	m_status=thinking;
	iterative_search();
	m_status=finished;
      }else if(m_action==stop_thinking){
	m_action=no_action;
	m_status=finished;
      }
      if(m_quit){break;}
      std::this_thread::sleep_for(milliseconds(5));
    }
}

const CounterMove& Analysis::countermove()const
{
  return m_cm;
}

const HistoryTable& Analysis::historytable()const
{
  return m_ht;
}

Evaluation& Analysis::evaluation()
{
  return m_eva;
}

void Analysis::iidsort(MoveList& ml,int d,int alpha,int beta)
{
  if(d<4)return;
  PrincipalVariation null_pv;
  //------------- ml[0] failed low ------------------
  for(size_t i=1;i<ml.size();i++){             
    make_move(ml[i]);
    m_iid_level++;
    int value=-pv_search(d/2-2,-beta,-alpha,null_pv);
    m_iid_level--;
    unmake_move(ml[i]);
    if(value>alpha){
      Move m=ml[i];
      for(size_t j=i;j>1;j--){
	ml[j]=ml[j-1];
      }
      ml[1]=m;
      break;
    }
  }
}

void Analysis::localsort(MoveList& ml,int depth,bool pvs,Move hm)
{
  int hm_v1{0},hm_v2{0},hm_v3{0},cm_v1{0},cm_v2{0},fm_v1{0},fm_v2{0};
  Move hm1,hm2,hm3,cm1,cm2,fm1,fm2;
  if(ml.size()<2){return;}

  
  // find best counter_move:
  if(cp.last_move()!=NullMove()){
    const Move& m1=cp.last_move();
    for(size_t i=0;i<ml.size();i++){
      Move m2=ml[i];
      int cmv=m_cm.value(m1,m2);
      if(m2.xcode()==xnormal_move){
	if(cmv>cm_v1){
	  cm_v2=cm_v1;
	  cm_v1=cmv;
	  cm2=cm1;
	  cm1=ml[i];
	}else if(cmv>cm_v2){
	  cm_v2=cmv;
	  cm2=m2;
	}
      }
    }
  }
  
  // find best follow-up move:
  size_t hs{cp.history().size()};
  if(hs>1 && cp.history()[hs-2].last_move!=NullMove()){
    const Move& m1=cp.history()[hs-2].last_move;
    for(size_t i=0;i<ml.size();i++){
      Move m2=ml[i];
      if(ml[i].xcode()==xnormal_move){
	if(m_cm.value(m1,m2)>fm_v1){
	  fm_v2=fm_v1;
	  fm_v1=m_cm.value(m1,m2);
	  fm2=fm1;
	  fm1=ml[i];
	}else if(m_cm.value(m1,m2)>fm_v2){
	  fm_v2=m_cm.value(m1,m2);
	  fm2=m2;
	}
      }
    }
  }
  
  // find best historical moves:
  for(size_t i=0;i<ml.size();i++){
    Move m2=ml[i];
    if(m2.xcode()==xnormal_move){
      if(m_ht.value(m2)>hm_v1){
	hm_v3=hm_v2;
	hm_v2=hm_v1;
	hm_v1=m_ht.value(m2);
	hm3=hm2;
	hm2=hm1;
	hm1=m2;
      }else if(m_ht.value(m2)>hm_v2){
	hm_v3=hm_v2;
	hm3=hm2;
	hm_v2=m_ht.value(m2);
	hm2=m2;
      }else if(m_ht.value(m2)>hm_v3){
	hm_v2=m_ht.value(m2);
	hm3=m2;
      }
    }
  }



  
  //------------------------------ set xcodes ------------------------------------
  for(unsigned i=0;i<ml.size();i++){
    if(ml[i]==hm){
      ml[i].xcode(hashtable_move);
    }else if(ml[i].capture()){
      if(ml[i].piece().abs_value()<=ml[i].capture().abs_value()){
	ml[i].xcode(good_capture);
      }else{
	if(see(ml[i])<0){
	  ml[i].xcode(losing_capture);
	}else{
	  ml[i].xcode(good_capture);
	}
      }
    }else if(ml[i].code()==promo_queen && see(ml[i])>=0){
      ml[i].xcode(promotion);
    }else if(ml[i]==m_killer.get_killerA(m_ply)){
      ml[i].xcode(killer);
    }else if(ml[i]==m_killer.get_killerB(m_ply)){
      ml[i].xcode(killer);
    }else if(ml[i]==cm1 || ml[i]==cm2){
      ml[i].xcode(counter_move);
    }else if(ml[i]==fm1 || ml[i]==fm2){
      ml[i].xcode(counter_move);
    }else if(ml[i]==hm1){
      ml[i].xcode(historical_move1);
    }else if(ml[i]==hm2){
      ml[i].xcode(historical_move2);
    }else if(ml[i]==hm3){
      ml[i].xcode(historical_move3);
    }else{
      if(cp.stm()==white){
	BitBoard bpa{(cp.black_pawn()>>9 & not_h_file) | (cp.black_pawn()>>7 & not_a_file)};
	if(ml[i].piece().abs_value()>pawn_value && (ml[i].to().bitboard() & bpa)){
	  ml[i].xcode(bad_move);
	}
      }else{
	BitBoard wpa{(cp.white_pawn()<<9 & not_a_file) | (cp.white_pawn()<<7 & not_h_file)};
	if(ml[i].piece().abs_value()>pawn_value && (ml[i].to().bitboard() & wpa)){
	  ml[i].xcode(bad_move);
	}
      }
    }
    
    if((pvs || m_ply<3) && depth>4 && ml[i].xcode()>hashtable_move){
      make_move(ml[i]);
      if(!cp.check()){
	int mating_value=mate_value-max_search_depth;
	int score=-nullmove_search(depth,-mating_value,-mating_value+1);
	if(score >= mating_value){
	  ml[i].xcode(mate_attack);
	}
      }
      unmake_move(ml[i]);
    }
  }
    
  //---------------------------------- sort --------------------------------------
  for(unsigned i=0;i<ml.size();i++){
    for(unsigned j=i+1;j<ml.size();j++){
      if(ml[i].xcode()>ml[j].xcode()){
	swap(ml[i],ml[j]);
      }
    }
  }
  //------------------------------------------------------------------------------
}


int Analysis::see(Square s2)
{
  Square s1;
  switch(cp.stm()){
  case black:s1=cp.black_weakest_attacker(s2);break;
  case white:s1=cp.white_weakest_attacker(s2);break;
  default:throw logic_error("Analysis::see   side to move is corrupt");
  }

  if(s1==no_square){return 0;}
  Move m=cp.convert_to_move(s1,s2);
  if(m.capture().king()){return infinite_value;}
  int capval=m.capture().abs_value();
  switch(m.code())
    {
    case normal_move:break;
    case promo_queen:capval+=queen_value-pawn_value;break;
    case promo_rook:capval+=rook_value-pawn_value;break;
    case promo_bishop:capval+=bishop_value-pawn_value;break;
    case promo_knight:capval+=knight_value-pawn_value;break;
    case castling:break;
    case en_passant:break;
    case nullmove: break;
    }
  cp.force_move(m);
  int value=capval-see(s2);
  cp.undo_forced_move(m);

  if(value>0){return value;}else{return 0;}
}

int Analysis::see(const Move& m)
{
  int value;
  int capval=m.capture().abs_value();
  switch(m.code())
    {
    case normal_move:break;
    case promo_queen:capval+=queen_value-pawn_value;break;
    case promo_rook:capval+=rook_value-pawn_value;break;
    case promo_bishop:capval+=bishop_value-pawn_value;break;
    case promo_knight:capval+=knight_value-pawn_value;break;
    case castling:break;
    case en_passant:break;
    case nullmove: break;
    }
  Square tof=m.to();
  cp.force_move(m);
  value=capval-see(tof);
  cp.undo_forced_move(m);
  return value;
}

int Analysis::quiescence(int alpha,int beta,int depth)
{
  if(depth<0){new_node();}
  uint32_t aml[256];
  MoveList ml(reinterpret_cast<Move*>(aml)); // (avoid Move constructor)
  array<uint64_t,64> bb_moves;
  m_mg.create_movetable(cp,bb_moves);
  
  int value;
  int standpat;
  if(!cp.check()){
    standpat=m_eva.static_eval();
  }else{
    standpat=-mate_value;
  }
  if(standpat>=mate_value-max_search_depth){standpat-=m_ply;}
  else if(standpat<=-mate_value+max_search_depth){standpat+=m_ply;}
  
  if(standpat>=beta){return beta;}
  if(alpha<standpat){alpha=standpat;}
  if(m_action==stop_thinking){return alpha;}

  if(cp.check()){
    m_mg.serialize_moves(cp,bb_moves,ml);
    if(ml.size()==0){return -mate_value+m_ply;}
  }else{
    m_mg.serialize_captures(cp,bb_moves,ml);
  }
  
  for(unsigned i=0;i<ml.size();i++){
    auto& m=ml[i];
    if(depth >= -8 || see(m)>0 || cp.check()){
      make_move(m);
      value=-quiescence(-beta,-alpha,depth-1);
      if(value>=mate_value-max_search_depth){value-=m_ply;}
      else if(value<=-mate_value+max_search_depth){value+=m_ply;}
      unmake_move(m);
      if(value>=beta){return beta;}
      if(value>alpha){alpha=value;}
      }
  }
  return alpha; 
}

int Analysis::static_eval()
{
  return m_eva.static_eval();
}

string Analysis::static_eval_report()
{
  static_eval();
  return m_eva.report();
}

string Analysis::dump_line()
{
  ostringstream oss;
  oss << "(";
  oss << cp.get_make_move_counter();
  oss << ") ";
  oss << cp.history_str();
  oss << endl;
  return oss.str();
}

void Analysis::put_in_tt(bound_t co,int depth,int score,Move best_move,unsigned char date)
{
  TTrecord ttr;
  ttr.set_bound(co);
  ttr.set_draft(depth);
  ttr.set_score(score,m_ply);
  ttr.set_move(best_move);
  ttr.set_date(date);
  ttr.set_key(cp.get_hash());
  m_tt.write(ttr);
}
