/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#ifndef ANALYZE_H
#define ANALYZE_H

#include<fstream>
#include<vector>
#include<utility>
#include<atomic>
#include<mutex>
#include"../movegen/mlist.h"
#include"../movegen/movegenerator.h"
#include"killer.h"
#include"principal.h"
#include"multipv.h"
#include"../eval/evaluation.h"
#include"../board/hasher.h"
#include"history.h"
#include"countermove.h"
#include"lmr.h"
#include"../board/rep3filter.h"
#include"timemanager.h"
#include"../common/tables.h"
using std::vector;
using std::pair;
using std::atomic;
using std::mutex;
using std::ofstream;



class ChessPosition;


enum action_t {no_action,start_iterations,stop_thinking};
enum status_t {idle,thinking,finished};


class Analysis
{  
 public:
  Analysis(ChessPosition&,Tables&);
  Analysis(const Analysis&,ChessPosition&);
  ~Analysis();
  void init(const ChessPosition&);

  void iterative_search();
  int root_search(MoveList&,int,int,int);
  int pv_search(const int&,int,int,PrincipalVariation&);
  int nullwindow_search(int,int,int);
  int nullmove_search(int,int,int);
  int extend(const MoveList&,const Move&);
  void iidsort(MoveList&,int,int,int);
  void localsort(MoveList&,int,bool,Move=NullMove());

  int see(Square);
  int see(const Move&);
  int quiescence(int,int,int);
  int static_eval();
  string static_eval_report();
  const CounterMove& countermove()const;
  const HistoryTable& historytable()const;
  Evaluation& evaluation();

  void clear();
  void clear_eval();
  void clear_hash();
  void clear_killer();
  void clear_history();
  void clear_node_count();
  void hashtable_enabled(bool);
  void set_lag(int);
  void set_log(bool);
  int eval()const;
  Move best_move()const;
  Move pondermove()const;
  void make_move(const Move&);
  void unmake_move(const Move&);
  void make_nullmove();
  void unmake_nullmove();
  bool nullmove_allowed()const;
  bool is_legal(TTmove&)const;
  void operator()();
  void set_multipv(unsigned int);
  void set_max_threads(unsigned int);
  void set_max_nodes(uint64_t);
  void set_id(int);
  void reset_max_nodes();
  void set_action(action_t);
  void set_status(int);
  int get_status()const;
  int get_id()const;
  uint64_t get_node_count()const;
  void add_node_count(uint64_t);
  uint64_t get_running_time()const;
  float get_ebf()const;
  int get_hashfull()const;
  int get_iteration()const;
  int get_seldepth()const;
  void set_aspiration(bool);
  void remove(const Move&,MoveList&);
  
  void quit();
  const ChessPosition& chessposition()const;
  PrincipalVariation pv()const;
  MultiPV mpv()const;
  string dump_line();

  //----------------- UCI commands: -----------------------
  //
  void reset();
  void set_wtime(int);
  void set_btime(int);
  void set_winc(int);
  void set_binc(int);
  void set_movestogo(int);
  void set_depth(int);
  void set_nodes(uint64_t);
  void set_mate(int);
  void set_movetime(int);
  void set_infinite(bool);
  void set_pondering(bool);
  void set_searchmoves(vector<Move>);
  void set_hash_size(int);
  void set_uci_limitstrength(bool);
  void set_uci_elo(int);
  void ponderhit();
  //
  //-------------------------------------------------------


 private:
  // other objects:
  ChessPosition& cp;
  Tables& m_tables;
  TranspositionTable& m_tt;
  MoveGenerator m_mg;
  Egt4Pieces m_egtkbnk;
  Egt4Pieces m_egtkqkr;
  const LmrTable& m_lmr;
  HistoryTable m_ht;
  CounterMove m_cm;
  Evaluation m_eva;
  PrincipalVariation m_pv;
  MultiPV m_mpv;
  Killer m_killer;
  TimeManager m_tman;
  Timer m_tm;

  // threadpool:
  int m_max_threads{1};
  
  // communication with engine:
  atomic<int> m_action;
  atomic<int> m_status;
  atomic<bool> m_quit;
  mutable mutex m_mutex_mpv;

  // coordinates:
  atomic<int> m_iteration{0};
  int m_ply{0};
  atomic<int> m_seldepth{0};
  unsigned char m_date{0};
  atomic<uint64_t> m_node_count{0};
  int m_nullmove_level{0};
  int m_iid_level{0};
  int m_mate_search_level{0};
  int m_id{0};

  // statistics:
  uint64_t m_leave_nodes{0};
  float m_ebf{0};

  // logging:
  ofstream ofs_log;
  bool m_log{false};

  // max values:
  int m_max_depth{max_search_depth-1};
  int m_sd_depth{max_search_depth-1};
  uint64_t m_max_nodes{infinite_nodes};

  // uci limits:
  vector<Move> m_searchmoves;
  bool m_uci_limitstrength{false};
  int m_uci_elo{2000};
  int m_elo_maxdepth{max_search_depth};
  int m_handicap{0};

  // result of analysis:
  int m_value;
  Move m_best_move;
  Move m_pondermove;

  // aspiration window:
  bool m_baspiration{true};
  
  // private functions:
  void select_searchmoves(MoveList&);
  int limitstrength(int)const;
  void new_node();
  void put_in_tt(bound_t,int,int,Move,unsigned char);
  
};


#endif
