/*

Copyright (C) 2019 Werner Taelemans, Belgium

This file is part of Galjoen.

Galjoen is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; version 3 of the License.

Galjoen is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with Galjoen; see the file COPYING.  If not, see
<http://www.gnu.org/licenses/>.

*/

#include"../engine/movegen/movegenerator.h"
#include"sanmove.h"
#include<iostream>
#include<stdexcept>
using std::cout;
using std::endl;
using std::logic_error;



SanMove::SanMove(const ChessPosition& x,MoveGenerator& y) : m_cp(x), m_mg(y)
{
  code(nullmove);
}

SanMove::SanMove(const SanMove& m) : m_cp(m.m_cp), m_mg(m.m_mg)
{
  m_move=m.m_move;
  text=m.text;
}

SanMove::SanMove(const ChessPosition& x, MoveGenerator& y,const Move& m) : m_cp(x), m_mg(y)
{
  *this=m;
}

SanMove::SanMove(const ChessPosition& x, MoveGenerator& y,const string& s) : m_cp(x), m_mg(y)
{
  *this=s;
}

SanMove& SanMove::operator=(const Move& m)
{
  code(m.code());
  check(m.check());
  piece(m.piece());
  capture(m.capture());
  from(m.from());
  to(m.to());

  string text_src_file,text_src_rank,text_dest_file,text_dest_rank,text_piece;
  string text_capture,text_promotion,text_check;
  
  if(m.piece().pawn()){text_piece="";}
  else if(m.piece().knight()){text_piece="N";}
  else if(m.piece().bishop()){text_piece="B";}
  else if(m.piece().rook()){text_piece="R";}
  else if(m.piece().queen()){text_piece="Q";}
  else if(m.piece().king()){text_piece="K";}
  if(m.capture().no_piece()){
    text_capture="";
  }else{
    text_capture="x";
    if(text_piece==""){text_src_file=m.from().file();}
  }
  text_dest_file=m.to().file();
  text_dest_rank=m.to().rank();  

  if(m.code()==promo_queen){text_promotion="=Q";}
  else if(m.code()==promo_rook){text_promotion="=R";}
  else if(m.code()==promo_bishop){text_promotion="=B";}
  else if(m.code()==promo_knight){text_promotion="=N";}

  BitBoard bb_to=m.to().bitboard();
  BitBoard bb_from;
  if(m.piece().white_piece()){
    bb_from=m_cp.white_attackers(bb_to);
  }else{
    bb_from=m_cp.black_attackers(bb_to);
  }

  if(bb_from.popcount()>1){
    bool disambiguate{false},same_file{false},same_rank{false};
    MoveList ml;
    m_mg.generate_moves(m_cp,ml);
    for(unsigned i=0;i<ml.size();i++)
      {
        auto& m3=ml[i];
	if(m==m3){continue;}
	if(m.piece()!=m3.piece()){continue;}
	if(m.to()!=m3.to()){continue;}
	if(m.from().file()==m3.from().file()){same_file=true;}
	if(from().rank()==m3.from().rank()){same_rank=true;}
	disambiguate=true;
      }
    if(disambiguate){
      if(same_file && !same_rank){
	text_src_rank=m.from().rank();
      }else if(same_file && same_rank){
	text_src_file=m.from().file();
	text_src_rank=m.from().rank();
      }else if(!same_file){
	text_src_file=m.from().file();
      }
    }    
  }
  
  if(m.check())
    {
      text_check="+";
      MoveList ml;
      m_mg.generate_moves(m_cp,ml);
      if(ml.size()==0){text_check="#";}
    }
  
  text=text_piece+text_src_file+text_src_rank+text_capture+text_dest_file+text_dest_rank+text_promotion;
  if(m.code()==castling){
    if(m.to()>m.from()){text="O-O";}else{text="O-O-O";}
  }
  text+=text_check;

  return *this;
}

SanMove& SanMove::operator=(string s)
{
  char sanpiece{-1},src_file{-1},src_rank{-1},dest_file{-1},dest_rank{-1};
  move_t sancode=normal_move;
  bool sancapture{false},sancheck{false};
  text=s;

  while(!s.empty() && (s.back()==' ' || s.back()=='!' || s.back()=='?')){s.pop_back();}
  if(s.back()=='+' || s.back()=='#'){sancheck=true;s.pop_back();} 

  if(s.back()==')'){throw logic_error("SanMove operator=(): Illegal character found");}
  if(s[s.size()-2]=='='){
    switch(s.back()){
    case 'N':sancode=promo_knight;break;
    case 'B':sancode=promo_bishop;break;
    case 'R':sancode=promo_rook;break;
    case 'Q':sancode=promo_queen;break;
    default:throw logic_error("SanMove operator=(): wrong promotion format");
    }
    s.pop_back();
    s.pop_back();
  }
  if(s.back()>='1' && s.back()<='8'){dest_rank=s.back();s.pop_back();}
  if(s.back()>='a' && s.back()<='h'){dest_file=s.back();s.pop_back();}
  if(s.size()>0 && s.back()=='x'){sancapture=true;s.pop_back();}
  if(s.size()>0 && s.back()>='1' && s.back()<='8'){src_rank=s.back();s.pop_back();}
  if(s.size()>0 && s.back()>='a' && s.back()<='h'){src_file=s.back();s.pop_back();}
  switch(text[0]){
  case 'N':sanpiece='N';break;
  case 'B':sanpiece='B';break;
  case 'R':sanpiece='R';break;
  case 'Q':sanpiece='Q';break;
  case 'K':sanpiece='K';break;
  default:sanpiece='P';
  }
  

  //-------------------------------------------------------------------------
  bool move_found=false;
  if(m_cp.stm()){
    switch(sanpiece){
    case 'P':piece(white_pawn_id);break;
    case 'N':piece(white_knight_id);break;
    case 'B':piece(white_bishop_id);break;
    case 'R':piece(white_rook_id);break;
    case 'Q':piece(white_queen_id);break;
    case 'K':piece(white_king_id);break;
    }
  }else{
    switch(sanpiece){
    case 'P':piece(black_pawn_id);break;
    case 'N':piece(black_knight_id);break;
    case 'B':piece(black_bishop_id);break;
    case 'R':piece(black_rook_id);break;
    case 'Q':piece(black_queen_id);break;
    case 'K':piece(black_king_id);break;
    }
  }
  

  //-------------------------------------------------------------------------
  MoveList ml;
  m_mg.generate_moves(m_cp,ml);
  for(unsigned i=0;i<ml.size();i++){
    auto& m=ml[i];
    if(text=="O-O"||text=="O-O+"||text=="O-O#"){
      if(m.code()==castling && m.to()>m.from()){
	from(m.from());
	to(m.to());
	code(m.code());
	check(m.check());
	piece(m.piece());
	capture(m.capture());
	move_found=true;
	break;
      }
    }else if(text=="O-O-O"||text=="O-O-O+"||text=="O-O-O#"){
      if(m.code()==castling && m.to()<m.from()){
	from(m.from());
	to(m.to());
	code(m.code());
	check(m.check());
	piece(m.piece());
	capture(m.capture());
	move_found=true;
	break;
      }
    }else if(sancode!=normal_move){
      if(m.piece()!=piece()){continue;}
      if(m.to().file()!=dest_file){continue;}
      if(m.to().rank()!=dest_rank){continue;}
      if(src_file!=-1 && m.from().file()!=src_file){continue;}
      if(src_rank!=-1 && m.from().rank()!=src_rank){continue;}
      from(m.from());
      to(m.to());
      code(sancode);
      check(m.check());
      piece(m.piece());
      capture(m.capture());
      move_found=true;
      break;
    }else{
      if(m.piece()!=piece()){continue;}
      if(m.to().file()!=dest_file){continue;}
      if(m.to().rank()!=dest_rank){continue;}
      if(src_file!=-1 && m.from().file()!=src_file){continue;}
      if(src_rank!=-1 && m.from().rank()!=src_rank){continue;}
      from(m.from());
      to(m.to());
      code(m.code());
      check(m.check());
      piece(m.piece());
      capture(m.capture());
      if(m_cp.get_piece_from(to()).no_piece() && sancapture){code(en_passant);}
      move_found=true;
      break;
    }
  }
  
  string E="SanMove operator=(string): illegal move \"";E+=text;E+="\"";
  E+="\nfen()=";E+=m_cp.fen();
  if(!move_found){throw logic_error(E);}
  
  check(sancheck);
  return *this;
}  


string SanMove::str()
{
  return text;
}
